# -*- coding: utf-8 -*-
from custodian.vasp import handlers as orig_handlers

from vise.custodian_extension.error_handlers import (
    ViseVaspErrorHandler, ViseUnconvergedErrorHandler, MemoryOverflowHandler,
    DivergingEnergyErrorHandler, TooLongTimeCalcErrorHandler)
from vise.util.logger import get_logger

# Note1: Custodian handler groups are modified to avoid IBRION=1, which does not
#        show efermi (& eigenvalues?) in vasprun.xml.
#        Therefore, IBRION=1 could be switched on only for rough calculations.
# Note2: Don't forget "()" to generate the ErrorHandler class object.

logger = get_logger(__name__)


def handler_group(name: str = "default",
                  nionic_steps: int = 1,
                  timeout: int = 518400) -> list:

    handlers = {
        "minimum":    [orig_handlers.MeshSymmetryErrorHandler(),
                       orig_handlers.PotimErrorHandler(),
                       orig_handlers.PositiveEnergyErrorHandler(),
                       orig_handlers.FrozenJobErrorHandler(),
                       orig_handlers.StdErrHandler(),
                       ViseVaspErrorHandler(),
                       ViseUnconvergedErrorHandler(),
                       MemoryOverflowHandler()],
        "default":    [orig_handlers.MeshSymmetryErrorHandler(),
                       orig_handlers.NonConvergingErrorHandler(),
                       orig_handlers.PotimErrorHandler(),
                       orig_handlers.PositiveEnergyErrorHandler(),
                       orig_handlers.FrozenJobErrorHandler(),
                       orig_handlers.StdErrHandler(),
                       ViseVaspErrorHandler(),
                       ViseUnconvergedErrorHandler(),
                       MemoryOverflowHandler(),
                       DivergingEnergyErrorHandler(),
                       TooLongTimeCalcErrorHandler(timeout=timeout),
                       ],
        "dielectric": [orig_handlers.MeshSymmetryErrorHandler(),
                       orig_handlers.NonConvergingErrorHandler(
                           nionic_steps=nionic_steps),
                       orig_handlers.PositiveEnergyErrorHandler(),
                       orig_handlers.FrozenJobErrorHandler(),
                       orig_handlers.StdErrHandler(),
                       orig_handlers.LrfCommutatorHandler(),
                       ViseUnconvergedErrorHandler(),
                       MemoryOverflowHandler(),
                       DivergingEnergyErrorHandler(),
                       TooLongTimeCalcErrorHandler(timeout=timeout),
                       ],
        "no_handler": []
    }

    try:
        return handlers[name]
    except KeyError:
        logger.critical(f"Accepted handler group names are  {handlers.keys()}")
        raise




# -*- coding: utf-8 -*-

from enum import unique, Enum


@unique
class Xc(Enum):
    """ Supported exchange-correlation treatment. """
    pbe = "pbe"
    pbesol = "pbesol"
    lda = "lda"
    scan = "scan"
    pbe0 = "pbe0"
    hse = "hse"

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s):

        for m in Xc:
            if m.value == s:
                return m
        if s == "perdew-zunger81":
            return Xc.lda
        raise AttributeError(f"Xc:{s} is not proper.\n "
                             f"Supported Xc:\n {cls.name_list()}")

    @classmethod
    def name_list(cls):
        return ', '.join([e.value for e in cls])

    @property
    def require_wavefunctions(self):
        return True if self in BEYOND_MGGA else False


LDA_OR_GGA = (Xc.pbe, Xc.pbesol, Xc.lda)
SEMILOCAL = (Xc.pbe, Xc.pbesol, Xc.lda, Xc.scan)
HYBRID_FUNCTIONAL = (Xc.pbe0, Xc.hse)
MGGA_OR_HYBRID = (Xc.pbe0, Xc.hse, Xc.scan)
BEYOND_MGGA = HYBRID_FUNCTIONAL



# -*- coding: utf-8 -*-
import subprocess
from collections import Counter
from copy import deepcopy

from custodian.custodian import ErrorHandler
from custodian.utils import backup
from custodian.vasp import handlers as orig_handlers
from pymatgen.io.vasp import Vasprun
from vise.input_set.incar import ViseIncar
from vise.input_set.vasp_input import ViseVaspInput
from vise.custodian_extension.modder import ViseVaspModder
from pymatgen.io.vasp.inputs import VaspInput, Incar, Kpoints
from pymatgen.io.vasp.outputs import Oszicar
from pymatgen.transformations.standard_transformations import \
    SupercellTransformation

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"

# Monkey patch
orig_handlers.Incar = ViseIncar
orig_handlers.VaspModder = ViseVaspModder
orig_handlers.VaspInput = ViseVaspInput


# When the original error handlers are modified, add *Vise* to the class name.
class ViseVaspErrorHandler(orig_handlers.VaspErrorHandler):

    error_msgs = deepcopy(orig_handlers.VaspErrorHandler.error_msgs)
    error_msgs["plane_wave_coeff"] = \
        ["ERROR: while reading WAVECAR, plane wave coefficients changed"]
    error_msgs.pop("brmix", None)

    def __init__(self,  output_filename="vasp.out", natoms_large_cell=100):
        super().__init__(output_filename=output_filename,
                         natoms_large_cell=natoms_large_cell,
                         errors_subset_to_catch=ViseVaspErrorHandler.error_msgs)

    def check(self):
        incar = Incar.from_file("INCAR")
        self.errors = set()
        errors = set()
        with open(self.output_filename, "r") as f:
            for line in f:
                l = line.strip()
                for err, msgs in self.error_msgs.items():
                    if err in self.errors_subset_to_catch:
                        for msg in msgs:
                            if l.find(msg) != -1:
                                # this checks if we want to run a charged
                                # computation (e.g., defects) if yes we don't
                                # want to kill it because there is a change in
                                # e-density (brmix error)
                                if err == "brmix" and 'NELECT' in incar:
                                    continue
                                self.errors.add(err)
                                errors.add(msg)
        for msg in errors:
            self.logger.error(msg, extra={"incar": incar.as_dict()})
        return len(self.errors) > 0

    def correct(self):

        backup(orig_handlers.VASP_BACKUP_FILES | {self.output_filename})
        actions = []
        vi = VaspInput.from_directory(".")

        if self.errors.intersection(["tet", "dentet"]):
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"ISMEAR": 0}}})

        if "inv_rot_mat" in self.errors:
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"SYMPREC": 1e-8}}})

        # ----- added ---------------------------------------------------
        if "plane_wave_coeff" in self.errors:
            actions.append({"file": "WAVECAR",
                            "action": {
                                "_file_delete": {'mode': "actual"}}})
            actions.append({"file": "CHGCAR",
                            "action": {
                                "_file_delete": {'mode': "actual"}}})
        # ---------------------------------------------------------------

        if "zpotrf" in self.errors:
            # Usually caused by short bond distances. If on the first step,
            # volume needs to be increased. Otherwise, it was due to a step
            # being too big and POTIM should be decreased.  If a static run
            # try turning off symmetry.
            try:
                oszicar = Oszicar("OSZICAR")
                nsteps = len(oszicar.ionic_steps)
            except:
                nsteps = 0

            if nsteps >= 1:
                potim = float(vi["INCAR"].get("POTIM", 0.5)) / 2.0
                actions.append(
                    {"dict": "INCAR",
                     "action": {"_set": {"ISYM": 0, "POTIM": potim}}})
            elif vi["INCAR"].get("NSW", 0) == 0 \
                    or vi["INCAR"].get("ISIF", 0) in range(3):
                actions.append(
                    {"dict": "INCAR", "action": {"_set": {"ISYM": 0}}})
            else:
                s = vi["POSCAR"].structure
                s.apply_strain(0.2)
                actions.append({"dict": "POSCAR",
                                "action": {
                                    "_set": {"structure": s.as_dict()}}})

            # Based on VASP forum's recommendation, you should delete the
            # CHGCAR and WAVECAR when dealing with this error.
            if vi["INCAR"].get("ICHARG", 0) < 10:
                actions.append({"file": "CHGCAR",
                                "action": {
                                    "_file_delete": {'mode': "actual"}}})
                actions.append({"file": "WAVECAR",
                                "action": {
                                    "_file_delete": {'mode': "actual"}}})

        if self.errors.intersection(["subspacematrix"]):
            if self.error_count["subspacematrix"] == 0:
                actions.append({"dict": "INCAR",
                                "action": {"_set": {"LREAL": False}}})
            else:
                actions.append({"dict": "INCAR",
                                "action": {"_set": {"PREC": "Accurate"}}})
            self.error_count["subspacematrix"] += 1

        if self.errors.intersection(
                ["rspher", "real_optlay", "nicht_konv"]):
            s = vi["POSCAR"].structure
            if len(s) < self.natoms_large_cell:
                actions.append({"dict": "INCAR",
                                "action": {"_set": {"LREAL": False}}})
            else:
                # for large supercell, try an in-between option LREAL = True
                # prior to LREAL = False
                if self.error_count['real_optlay'] == 0:
                    # use real space projectors generated by pot
                    actions.append({"dict": "INCAR",
                                    "action": {"_set": {"LREAL": True}}})
                elif self.error_count['real_optlay'] == 1:
                    actions.append({"dict": "INCAR",
                                    "action": {"_set": {"LREAL": False}}})
                self.error_count['real_optlay'] += 1

        if self.errors.intersection(["tetirr", "incorrect_shift"]):

            # --Modified------------------------------------------------------
            if vi["KPOINTS"].style == Kpoints.supported_modes.Monkhorst or \
                    vi["KPOINTS"].kpts_shift != [0.0, 0.0, 0.0]:
                actions.append({"dict": "KPOINTS",
                                "action": {
                                    "_set": {"generation_style": "Gamma",
                                             "usershift": [0.0, 0.0,
                                                           0.0]}}})
            # -----------------------------------------------------------
        if "rot_matrix" in self.errors:
            # --Modified------------------------------------------------------
            if vi["KPOINTS"].style == Kpoints.supported_modes.Monkhorst or \
                    vi["KPOINTS"].kpts_shift != [0.0, 0.0, 0.0]:
                actions.append({"dict": "KPOINTS",
                                "action": {
                                    "_set": {"generation_style": "Gamma",
                                             "usershift": [0.0, 0.0,
                                                           0.0]}}})
            # -----------------------------------------------------------
            else:
                actions.append({"dict": "INCAR",
                                "action": {"_set": {"ISYM": 0}}})

        if "amin" in self.errors:
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"AMIN": "0.01"}}})

        if "triple_product" in self.errors:
            s = vi["POSCAR"].structure
            trans = SupercellTransformation(
                ((1, 0, 0), (0, 0, 1), (0, 1, 0)))
            new_s = trans.apply_transformation(s)
            actions.append({"dict": "POSCAR",
                            "action": {
                                "_set": {"structure": new_s.as_dict()}},
                            "transformation": trans.as_dict()})

        if "pricel" in self.errors:
            actions.append({"dict": "INCAR",
                            "action": {
                                "_set": {"SYMPREC": 1e-8, "ISYM": 0}}})

        if "brions" in self.errors:
            potim = float(vi["INCAR"].get("POTIM", 0.5)) + 0.1
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"POTIM": potim}}})

        if "zbrent" in self.errors:
            # Modified so as not to use IBRION=1 as it does not show the
            # eigenvalues in vasprun.xml >>>>>>>>>>>>
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"ADDGRID": True}}})
            actions.append({"file": "CONTCAR",
                            "action": {"_file_copy": {"dest": "POSCAR"}}})
        #            actions.append({"dict": "INCAR",
        #                            "action": {"_set": {"IBRION": 1}}})
        #            actions.append({"file": "CONTCAR",
        #                            "action": {"_file_copy": {"dest": "POSCAR"}}})
        # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

        if "too_few_bands" in self.errors:
            if "NBANDS" in vi["INCAR"]:
                nbands = int(vi["INCAR"]["NBANDS"])
            else:
                with open("OUTCAR") as f:
                    for line in f:
                        if "NBANDS" in line:
                            try:
                                d = line.split("=")
                                nbands = int(d[-1].strip())
                                break
                            except (IndexError, ValueError):
                                pass

        # >>>>>>>>>>>>
        # Modified since when the nbands is less than 9, it is not incremented
            actions.append({"dict": "INCAR",
                            "action": {
                                "_set": {"NBANDS": int(1.1 * nbands) + 2}}})
                   #            "_set": {"NBANDS": int(1.1 * nbands)}}})
        # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        if "pssyevx" in self.errors:
            actions.append({"dict": "INCAR", "action":
                {"_set": {"ALGO": "Normal"}}})
        if "eddrmm" in self.errors:
            # RMM algorithm is not stable for this calculation
            if vi["INCAR"].get("ALGO", "Normal") in ["Fast", "VeryFast"]:
                actions.append({"dict": "INCAR", "action":
                    {"_set": {"ALGO": "Normal"}}})
            else:
                potim = float(vi["INCAR"].get("POTIM", 0.5)) / 2.0
                actions.append({"dict": "INCAR",
                                "action": {"_set": {"POTIM": potim}}})
            if vi["INCAR"].get("ICHARG", 0) < 10:
                actions.append({"file": "CHGCAR",
                                "action": {
                                    "_file_delete": {'mode': "actual"}}})
                actions.append({"file": "WAVECAR",
                                "action": {
                                    "_file_delete": {'mode': "actual"}}})

        if "edddav" in self.errors:
            if vi["INCAR"].get("ICHARG", 0) < 10:
                actions.append({"file": "CHGCAR",
                                "action": {
                                    "_file_delete": {'mode': "actual"}}})
            actions.append({"dict": "INCAR", "action":
                {"_set": {"ALGO": "All"}}})

        if "grad_not_orth" in self.errors:
            if vi["INCAR"].get("ISMEAR", 1) < 0:
                actions.append({"dict": "INCAR",
                                "action": {"_set": {"ISMEAR": "0"}}})

        if "zheev" in self.errors:
            if vi["INCAR"].get("ALGO", "Fast").lower() != "exact":
                actions.append({"dict": "INCAR",
                                "action": {"_set": {"ALGO": "Exact"}}})
        if "elf_kpar" in self.errors:
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"KPAR": 1}}})

        if "rhosyg" in self.errors:
            if vi["INCAR"].get("SYMPREC", 1e-4) == 1e-4:
                actions.append({"dict": "INCAR",
                                "action": {"_set": {"ISYM": 0}}})
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"SYMPREC": 1e-4}}})

        if "posmap" in self.errors:
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"SYMPREC": 1e-6}}})

        if "point_group" in self.errors:
            actions.append({"dict": "INCAR",
                            "action": {"_set": {"ISYM": 0}}})

        ViseVaspModder(vi=vi).apply_actions(actions)
        return {"errors": list(self.errors), "actions": actions}


class ViseUnconvergedErrorHandler(orig_handlers.UnconvergedErrorHandler):

    def correct(self):
        backup(orig_handlers.VASP_BACKUP_FILES)
        v = Vasprun(self.output_filename)
        actions = [{"file":   "CONTCAR",
                    "action": {"_file_copy": {"dest": "POSCAR"}}}]
        if not v.converged_electronic:
            # For SCAN try switching to CG for the electronic minimization
            if "SCAN" in v.incar.get("METAGGA", "").upper():
                new_settings = {"ALGO": "All"}
            else:
                new_settings = {"ISTART":   1,
                                "ALGO":     "Normal",
                                "NELMDL":   -6,
                                "BMIX":     0.001,
                                "AMIX_MAG": 0.8,
                                "BMIX_MAG": 0.001}

            if all([v.incar.get(k, "") == val for k, val in
                    new_settings.items()]):
                return {"errors": ["Unconverged"], "actions": None}

            actions.append({"dict":   "INCAR",
                            "action": {"_set": new_settings}})

        # Instead of changing the IBRION, we reduce the EDIFF since we usually
        # use the constant EDIFF value as default.
        elif not v.converged_ionic:
            ediff = v.incar["EDIFF"]
            actions.append({"dict":   "INCAR",
                            "action": {"_set": {"EDIFF": ediff * 0.5}}})
            actions.append({"dict":   "INCAR",
                            "action": {"_set": {"ADDGRID": True}}})
        ViseVaspModder().apply_actions(actions)
        return {"errors": ["Unconverged"], "actions": actions}


class DielectricMaxIterationErrorHandler(ErrorHandler):
    """Detects if the SCF is not attained. """

    is_monitor = True

    def __init__(self, oszicar="OSZICAR", incar="INCAR"):
        self.error_count = Counter()
        self.oszicar = oszicar
        self.incar = incar

    def check(self):

        incar = Incar.from_file(self.incar)
        nelm = incar.get("NELM", 60)
        unconverged_line = f":{nelm:4d}"
        with open(self.oszicar, "r") as fr:
            return any(unconverged_line in line for line in fr)

    def correct(self):
        # Uncorrectable error. Just return None for actions.
        return {"errors": ["No_DFPT_convergence"], "actions": None}


class MemoryOverflowHandler(ErrorHandler):
    """Detects if the memory is overflowed. """

    is_monitor = True

    def __init__(self, memory_usage_limit=0.85):
        """
        Initializes the handler with the output file to check.

        Args:
            memory_usage_limit (float):
        """
        self.memory_usage_limit = memory_usage_limit

    def check(self):
        res = subprocess.check_output('free').split()
        memory_usage = int(res[8]) / int(res[7])
        if memory_usage > self.memory_usage_limit:
            return True

    def correct(self):
        # Uncorrectable error. Just return None for actions.
        return {"errors": ["Too_much_memory_usage"], "actions": None}


class TooLongTimeCalcErrorHandler(ErrorHandler):
    """Detects if calculation runtime is longer than the given timeout."""

    is_monitor = True

    def __init__(self, timeout=129600):
        """

        60 * 60 * 36 = 129600 (36 hours)

        Args:
            timeout (int):
        """
        self.timeout = timeout

    def check(self):
        now_time = int(subprocess.check_output(['date', '+%s']))
        incar_time = \
            int(subprocess.check_output(['date', '+%s', '-r', 'INCAR']))
        incar_age = now_time - incar_time
        if incar_age > self.timeout:
            return True

    def correct(self):
        # Uncorrectable error. Just return None for actions.
        return {"errors": ["Too_long_calc"], "actions": None}


class DivergingEnergyErrorHandler(ErrorHandler):

    def __init__(self, output_filename="OSZICAR", energy_criterion=10**6):
        """Initializes the handler with the output file to check.

        Args:
            output_filename (str): This is the OSZICAR file. Change
                this only if it is different from the default (unlikely).
        """
        self.output_filename = output_filename
        self.energy_criterion = energy_criterion

    def check(self):
        oszicar = Oszicar(self.output_filename)
        esteps = oszicar.electronic_steps
        # OSZICAR file can be empty, thus we need try-except here.
        try:
            max_energy = max([es["E"] for es in esteps[-1]])
        except (IndexError, KeyError):
            return False

        if max_energy > self.energy_criterion:
            return True

    def correct(self):
        # Uncorrectable error. Just return None for actions.
        return {"errors": ["Energy_diverging"], "actions": None}


class ReturnErrorHandler(ErrorHandler):
    """Return Error for test."""
    is_monitor = True

    def __init__(self):
        pass

    def check(self):
        return True

    def correct(self):
        # Uncorrectable error. Just return None for actions.
        return {"errors": ["Always return Error with this."], "actions": None}






# -*- coding: utf-8 -*-

import numpy as np

from pymatgen.io.vasp import Kpoints
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice

from vise.input_set.make_kpoints import (
    KpointsMode, MakeKpoints, irreducible_kpoints, num_irreducible_kpoints)
from vise.util.testing import ViseTest

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class KpointsModeTest(ViseTest):
    def test(self):
        band = KpointsMode.from_string("band")
        self.assertEqual(KpointsMode.band, band)

    def test2(self):
        with self.assertRaises(AttributeError):
            KpointsMode.from_string("fail_string")


class MakeKpointsTest(ViseTest):

    def setUp(self) -> None:
        self.sg1 = self.get_structure_by_sg(1)  # test for hinuma reduced cell
        self.sg23 = self.get_structure_by_sg(23)  # I222, body-centered ortho
#        self.sg38 = self.get_structure_by_sg(38)  # Amm2, base-centered ortho

    def test_band_sg1(self):
        k = MakeKpoints(mode="band",
                        only_even=True,
                        structure=self.sg1)
        k.make_kpoints()
        expected_kpt_mesh = [2, 4, 4]
        expected_lattice = Lattice([[1.702429, 1.411278, - 11.284670],
                                    [1.502067, -6.074800, 0.000000],
                                    [-6.169335, 0.000000, 0.000000]])
        expected_shift = [0.5, 0.5, 0.5]
        self.assertEqual(expected_kpt_mesh, k.kpt_mesh)
        self.assertEqual(expected_lattice, k.corresponding_structure.lattice)
        self.assertEqual(expected_shift, k.kpt_shift)
        self.assertEqual(161, k.num_kpts)

    def test_band_sg23(self):
        k = MakeKpoints(mode="band",
                        only_even=True,
                        structure=self.sg23)
        k.make_kpoints()
        expected_kpt_mesh = [4, 4, 4]
        expected_lattice = Lattice([
            [-3.5435465044428671, 3.7004010644681178, 4.1207036549338687],
            [3.5435465044428671, -3.7004010644681178, 4.1207036549338687],
            [3.5435465044428671, 3.7004010644681178, -4.1207036549338687]])
        expected_shift = [0.5, 0.5, 0.5]
        self.assertEqual(expected_kpt_mesh, k.kpt_mesh)
        self.assertEqual(expected_lattice, k.corresponding_structure.lattice)
        self.assertEqual(expected_shift, k.kpt_shift)
        self.assertEqual(224, k.num_kpts)

    def test_primitive_uniform_sg1(self):
        k = MakeKpoints(mode="primitive_uniform",
                        structure=self.sg1)
        k.make_kpoints()
        expected_kpt_mesh = [3, 3, 2]
        expected_lattice = Lattice([[6.169335, 0.000000, 0.000000],
                                    [-1.502067, 6.074800, 0.000000],
                                    [-1.702429, -1.411278, 11.284670]])

        expected_shift = [0.0, 0.0, 0.5]
        self.assertEqual(expected_kpt_mesh, k.kpt_mesh)
        self.assertEqual(expected_lattice, k.corresponding_structure.lattice)
        self.assertEqual(expected_shift, k.kpt_shift)
        self.assertEqual(9, k.num_kpts)

    def test_manual_set_sg1(self):
        k = MakeKpoints(mode="manual_set",
                        structure=self.sg1)
        k.make_kpoints()
        expected_kpt_mesh = [3, 3, 2]
        expected_lattice = Lattice([[6.16933470, 0.00090247, -0.00277192],
                                    [-1.5029601, 6.07456979, -0.01071425],
                                    [-1.6971538, -1.3903666, 11.28805959]])

        expected_shift = [0.0, 0.0, 0.0]
        self.assertEqual(expected_kpt_mesh, k.kpt_mesh)
        self.assertEqual(expected_lattice, k.corresponding_structure.lattice)
        self.assertEqual(expected_shift, k.kpt_shift)
        self.assertEqual(10, k.num_kpts)


class IrreducibleKpointsTest(ViseTest):
    def setUp(self):
        lattice = [[10, 0, 0], [-5, 8.660254, 0], [0, 0, 5]]
        self.hexagonal = Structure(lattice=lattice,
                                   species=["H"],
                                   coords=[[0, 0, 0]])
        self.kpoints_g = Kpoints.gamma_automatic(kpts=(2, 2, 2))

        lattice2 = [[1, 0, 0], [0, 2, 0], [0, 0, 3]]
        self.ortho = Structure(lattice=lattice2,
                               species=["H"],
                               coords=[[0, 0, 0]])
        self.kpoints_mp = Kpoints.monkhorst_automatic(kpts=(2, 2, 2))

        self.kpoints_reciprocal = \
            Kpoints(style=Kpoints.supported_modes.Reciprocal,
                    num_kpts=2,
                    kpts=((0, 0, 0), (0.5, 0.5, 0.5)),
                    kpts_weights=[1, 1])

    def test_gamma(self):
        actual = irreducible_kpoints(structure=self.hexagonal,
                                     kpoints=self.kpoints_g)
        expected = [(np.array([0.0,  0.0,  0.0]), 1),
                    (np.array([0.5,  0.0,  0.0]), 3),
                    (np.array([0.0,  0.0,  0.5]), 1),
                    (np.array([0.5,  0.0,  0.5]), 3)]
        for i in range(4):
            self.assertEqual(expected[i][0].tolist(), actual[i][0].tolist())
            self.assertEqual(expected[i][1], actual[i][1])

    def test_mp(self):
        actual = irreducible_kpoints(structure=self.ortho,
                                     kpoints=self.kpoints_mp)
        expected = [(np.array([0.25,  0.25,  0.25]), 8)]
        self.assertEqual(expected[0][0].tolist(), actual[0][0].tolist())
        self.assertEqual(expected[0][1], actual[0][1])

    def test_num_kpts(self):
        self.assertEqual(2, num_irreducible_kpoints(self.kpoints_reciprocal))
        self.assertEqual(4, num_irreducible_kpoints(structure=self.hexagonal,
                                                    kpoints=self.kpoints_g))
        self.assertEqual(1, num_irreducible_kpoints(structure=self.ortho,
                                                    kpoints=self.kpoints_mp))



# -*- coding: utf-8 -*-
import logging
import sys

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


def get_logger(name: str,
               level=logging.DEBUG,
               stream=sys.stdout,
               log_format: str = None,
               log_filename: str = None):
#    log_filename: str = "vise.log"):

    log_format = log_format or \
                 '%(asctime)25s %(levelname)7s %(name)25s \n --> %(message)s'

    logger = logging.getLogger(name)
    logger.setLevel(level)

    formatter = logging.Formatter(log_format)

    if log_filename:
        fh = logging.FileHandler(log_filename)
        fh.setFormatter(formatter)
        fh.setLevel(logging.WARNING)
        logger.addHandler(fh)

    sh = logging.StreamHandler(stream=stream)
    sh.setFormatter(formatter)
    logger.addHandler(sh)

    return logger



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np

from typing import Tuple, Optional, Union
from pymatgen.io.vasp.outputs import Vasprun, Outcar
from pymatgen.electronic_structure.core import Spin


def band_gap_properties(vasprun: Union[Vasprun, str],
                        outcar: Union[Outcar, str],
                        int_m_threshold: float = 0.1
                        ) -> Tuple[dict, Optional[dict], Optional[dict]]:
    """Evaluate the band gap properties from vasprun.xml and OUTCAR files

    Args:
        vasprun (Vasprun/ str):
            Vasprun file.
        outcar (Outcar/ str):
           OUTCAR file.
        int_m_threshold:
            The threshold to judge if the magnetization is not integer.

    Return:
        Tuple of band_gap, vbm, and cbm-related info. For metals,
        {'energy': 0.0, 'direct': False, 'transition': None}, None, None
        is returned.
    """
    if isinstance(vasprun, str):
        vasprun = Vasprun(vasprun)
    if isinstance(outcar, str):
        outcar = Outcar(outcar)

    eigenvalues = vasprun.eigenvalues
    kpts = vasprun.actual_kpoints
    metal = {'energy': 0.0, 'direct': False, 'transition': None}, None, None
    mag = outcar.total_mag

    if mag is None:
        if abs(outcar.nelect - round(outcar.nelect / 2) * 2) > int_m_threshold:
            return metal
        hob_index = round((outcar.nelect / 2) - 1)

        hob_eig, hob_k_index, lub_eig, lub_k_index = \
            edge_info(eigenvalues, hob_index)

        vbm_info = {'energy': hob_eig,
                    'spin': None,
                    'band_index': hob_index,
                    'kpoints': kpts[hob_k_index]}
        cbm_info = {'energy': lub_eig,
                    'spin': None,
                    'band_index': hob_index + 1,
                    'kpoints': kpts[lub_k_index]}

    else:
        frac_occupation = mag - round(mag)
        if abs(frac_occupation) > int_m_threshold:
            return metal

        up_hob_index = round((outcar.nelect + (mag + 0.1)) / 2) - 1
        up_hob_eig, up_hob_k_index, up_lub_eig, up_lub_k_index = \
            edge_info(eigenvalues, up_hob_index)

        down_hob_index = round((outcar.nelect - (mag + 0.1)) / 2) - 1
        down_hob_eig, down_hob_k_index, down_lub_eig, down_lub_k_index = \
            edge_info(eigenvalues, down_hob_index, Spin.down)

        if up_hob_eig >= down_hob_eig:
            vbm_info = {'energy': up_hob_eig,
                        'spin': int(Spin.up),
                        'band_index': up_hob_index,
                        'kpoints': kpts[up_hob_k_index]}
        else:
            vbm_info = {'energy': down_hob_eig,
                        'spin': int(Spin.down),
                        'band_index': down_hob_index,
                        'kpoints': kpts[down_hob_k_index]}

        if up_lub_eig <= down_lub_eig:
            cbm_info = {'energy': up_lub_eig,
                        'spin': int(Spin.up),
                        'band_index': up_hob_index + 1,
                        'kpoints': kpts[up_lub_k_index]}
        else:
            cbm_info = {'energy': down_lub_eig,
                        'spin': int(Spin.down),
                        'band_index': down_hob_index + 1,
                        'kpoints': kpts[down_lub_k_index]}

    if vbm_info["energy"] > cbm_info["energy"]:
        return metal

    is_direct = (vbm_info["spin"] == cbm_info["spin"]
                 and vbm_info["kpoints"] == cbm_info["kpoints"])

    band_gap = {"energy": cbm_info["energy"] - vbm_info["energy"],
                "direct": is_direct}

    return band_gap, vbm_info, cbm_info


def edge_info(eigenvalues: dict,
              hob_index: int,
              spin=Spin.up
              ) -> Tuple[float, int, float, int]:
    """Evaluate the band edge info from eigenvalues and given band index.

    Args:
        eigenvalues (dict):
            Dict of eigenvalues.
            eigenvalue, occupation =
            Eigenvalues[spin][k-point index][band index][0:2]
        hob_index (int):
            Highest-occupied band (HOB) index starting from0.
        spin:
            Pymatgen Spin object.

    Return:
        Tuple of HOB eigenvalue, its k index, lowest-unoccupied band (LUB)
        eigenvalue, and its k index.
    """

    hob_eig = np.amax(eigenvalues[spin][:, hob_index, 0])
    hob_k_index = int(np.where(
        eigenvalues[spin][:, hob_index, 0] == hob_eig)[0][0])

    lub_eig = np.amin(eigenvalues[spin][:, hob_index + 1, 0])
    lub_k_index = int(np.where(
        eigenvalues[spin][:, hob_index + 1, 0] == lub_eig)[0][0])

    return hob_eig, hob_k_index, lub_eig, lub_k_index




# -*- coding: utf-8 -*-

import os
from pymatgen.electronic_structure.core import Spin
from pathlib import Path

from vise.analyzer.dos_plotter import get_dos_plot, max_density
from vise.util.testing import ViseTest

parent_dir = Path(__file__).parent


class ViseDosPlotterTest(ViseTest):

    # Must split pyplot constructors. Otherwise, mgo_dos.pdf shows NaN3 results.
    def test_normal(self):
        mgo_dos = get_dos_plot(vasprun=parent_dir / "MgO_dos_vasprun.xml")
        mgo_dos.savefig(str(parent_dir / "mgo_dos.pdf"))

    def test_w_args(self):
        nan3_dos_mod = get_dos_plot(vasprun=parent_dir / "NaN3_dos_vasprun.xml",
                                    pdos_type="none",
                                    orbital=False,
                                    xlim=[-3, 5],
                                    ymaxs=[10, 5],
                                    zero_at_efermi=False,
                                    specific=["1"],
                                    crop_first_value=False,
                                    show_spg=False,
                                    )
        nan3_dos_mod.savefig("nan3_dos.pdf")
        os.remove("nan3_dos.pdf")


class MaxDensityTest(ViseTest):
    def setUp(self) -> None:
        self.density = {Spin.up: [10.0, 1.0, 2.0, 3.0],
                        Spin.down: [4.0, 8.0, 6.0, 7.0]}
        self.energies = [0.0, 10.0, 20.0, 230.0]
        self.xlim = [5.0, 25.0]
        self.crop_first_value = True

    def test(self):
        expected = 8.0
        actual = max_density(self.density,
                             self.energies,
                             self.xlim,
                             self.crop_first_value)
        self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


SYMMETRY_TOLERANCE = 0.01
ANGLE_TOL = 5

DOS_STEP_SIZE = 0.01

KPT_DENSITY = 2.5

KPT_INIT_DENSITY = 2.5  # Initial k-point density used for k-point convergence
KPT_FACTOR = 1.2  # Multiplied factor used for incrementing the k-point density.

BAND_REF_DIST = 0.025

ENCUT_FACTOR_STR_OPT = 1.3  # This times ENMAX is used for structure opt calc.

BAND_GAP_CRITERION = 0.2  # Criterion in eV to determine if band gap exists.

DEFAULT_NUM_NODES = 1

TIMEOUT = 518400

# chempotdiag
ROOM_TEMPERATURE = 298.15
REFERENCE_PRESSURE = 1e+5
MOLECULE_SUFFIX = "molecule_chempotdiag"



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from copy import deepcopy
from pathlib import Path
from typing import List, Dict, Union
import re

from monty.json import MSONable

from pymatgen.analysis.phase_diagram import PDEntry
from pymatgen.core.composition import Composition
from pymatgen.entries.entry_tools import EntrySet
from pymatgen.io.vasp import Vasprun

from vise.util.mp_tools import get_mp_materials
from vise.chempotdiag.gas import Gas
from vise.config import REFERENCE_PRESSURE
from vise.util.logger import get_logger
from vise.util.tools import parse_file

logger = get_logger(__name__)


class FreeEnergyEntry(PDEntry):

    def __init__(self,
                 composition: Union[Composition, str],
                 total_energy: float,
                 zero_point_vib: float = None,
                 free_e_shift: float = None,
                 name: str = None,
                 data: dict = None,
                 attribute: object = None):
        """Free energy version of PDEntry with zero_point_vib and free_e_shift

        Name is used for plotting chemical potential and phase diagrams. It is
        especially important for CompoundPhaseDiagram, as the composition is
        described with DummySpecie.

        Args:
            composition (Composition/str):
                Composition of the entry.
            total_energy (float):
                Energy of the entry. Usually final calculated energy from VASP.
            zero_point_vib (float):
                Zero point vibration energy.
            free_e_shift (float):
                Entropic contribution to the free energy
            name (str):
                Name of the Entry. Mostly used for drawing the figure.
            data (dict):
                An optional dict of any additional data associated
                with the entry. Defaults to None.
            attribute:
                Optional attribute of the entry. This can be used to specify
                that the entry is a newly found compound, or to specify a
                particular label for the entry, or else ... Used for further
                analysis and plotting purposes. An attribute can be anything
                but must be MSONable.
        """
        super().__init__(composition, 0.0, name, attribute)
        self.total_energy = total_energy
        self.zero_point_vib = zero_point_vib
        self.free_e_shift = free_e_shift
        self.data = data
        self.energy = total_energy

        if zero_point_vib:
            self.energy += zero_point_vib
        if free_e_shift:
            self.energy += free_e_shift

    def as_dict(self):
        return {"composition": self.composition.as_dict(),
                "total_energy": self.total_energy,
                "zero_point_vib": self.zero_point_vib,
                "free_e_shift": self.free_e_shift,
                "name": self.name,
                "data": self.data,
                "attribute": self.attribute}

    @classmethod
    def from_dict(cls, d: dict):
        kwargs = deepcopy(d)
        kwargs["composition"] = Composition(d["composition"])
        kwargs.pop("@module", None)
        kwargs.pop("@class", None)
        return cls(**kwargs)

    def __repr__(self):
        zero_point_vib = \
            f"{self.zero_point_vib:.3f}" if self.zero_point_vib else "none"
        free_e_shift = \
            f"{self.free_e_shift:.3f}" if self.free_e_shift else "none"
        return f"PDEntry : {self.name} with composition {self.composition}. " \
               f"Energy: {self.energy:.3f}. " \
               f"Zero point vib energy: {zero_point_vib}. " \
               f"Free energy contribution: {free_e_shift}. " \
               f"Data: {self.data}"


class FreeEnergyEntrySet(EntrySet, MSONable):
    """List of FreeEnergyEntries with temperature and pressure."""
    def __init__(self,
                 entries: List[FreeEnergyEntry],
                 temperature: float = None,
                 pressure: float = None):
        super().__init__(entries)
        self.temperature = temperature
        self.pressure = pressure

    @classmethod
    def from_vasp_files(cls,
                        directory_paths: List[Path],
                        vasprun: str = "vasprun.xml",
                        parse_gas: bool = True,
                        temperature: float = 0.0,
                        partial_pressures: Dict[str, float] = None
                        ) -> "FreeEnergyEntrySet":
        """ Constructs class object from vasp output files.

        Args:
            directory_paths (list):
                List of relative directory paths, containig vasp results.
            vasprun (str):
                Name of the vasprun.xml file.
            parse_gas (bool):
                Whether to parse gases as gas phases if the composition exists
                in the Gas.name_list().
            temperature (float):
                Temperature in K.
            partial_pressures (dict):
                Dict of species as keys (str) and pressures in Pa as values.
                Example: {"O2": 2e5, "N2": 70000}

        Returns:
            FreeEnergyEntrySet class object.
        """
        energy_entries = []
        for d in directory_paths:
            logger.info(f"Parsing data in {d} ...")
            v: Vasprun = parse_file(Vasprun, Path(d) / vasprun)
            composition = v.final_structure.composition.formula

            kwargs = {}
            mol_dir = re.match(r"^mol_", str(d))
            if parse_gas and mol_dir:
                if composition not in Gas.name_list():
                    logger.critical(f"{d} does not exist in Gas.name_list, so "
                                    f"skipp to generate the zero point vib "
                                    f"and free energies.")

                if partial_pressures is None:
                    logger.info(f"Pressure of {composition} is set to "
                                f"{REFERENCE_PRESSURE} (Pa).")
                    pressure = REFERENCE_PRESSURE
                else:
                    try:
                        pressure = partial_pressures[composition]
                    except KeyError as e:
                        msg = f"Partial pressure for {composition} is not set."
                        raise Exception(msg) from e
                        
                gas = Gas[composition]
                zpve = gas.zero_point_vibrational_energy
                free_e = gas.free_e_shift(temperature, pressure)
                logger.warning(
                    f"{composition} is parsed as gas. "
                    f"Pressure: {pressure}, "
                    f"temperature: {temperature}, "
                    f"zero point vib energy: {zpve}, "
                    f"free energy: {free_e}")

                kwargs = {"zero_point_vib": zpve, "free_e_shift": free_e}

            energy_entries.append(
                FreeEnergyEntry(composition=composition,
                                total_energy=v.final_energy, **kwargs))

        return cls(energy_entries, temperature, partial_pressures)

    @classmethod
    def from_mp(cls, elements: List[str]) -> "FreeEnergyEntrySet":
        """Obtain the energies from Materials Project."""
        properties = ["task_id", "full_formula", "final_energy"]
        materials = get_mp_materials(elements, properties=properties)

        energy_entries = []
        for m in materials:
            energy_entries.append(
                FreeEnergyEntry(composition=m["full_formula"],
                                total_energy=m["final_energy"],
                                data={"mp_id": m["task_id"]}))

        return cls(energy_entries)

    def __repr__(self):
        lines = [f"pressure {self.pressure}",
                 f"temperature {self.temperature}"]
        lines.extend([str(i) for i in self.entries])

        return "\n".join(lines)


class ConstrainedFreeEnergyEntrySet(FreeEnergyEntrySet):
    def __init__(self,
                 entries: List[FreeEnergyEntry],
                 original_entries:  List[FreeEnergyEntry],
                 constr_chempot: Dict[str, float],
                 temperature: float = None,
                 pressure: float = None):
        """FreeEnergyEntrySet obtained at the constrained chemical potential.

       Args:
           entries (List[FreeEnergyEntry]):
               Entries for the
           original_entries (List[FreeEnergyEntry]):
           constr_chempot (Dict[str, float]):
           temperature (float):

           pressure (float):

        """
        super().__init__(entries, temperature, pressure)
        self.original_entries = original_entries
        self.constr_chempot = constr_chempot

    @classmethod
    def from_entry_set(cls,
                       entry_set: FreeEnergyEntrySet,
                       constr_chempot: Dict[str, float]
                       ) -> "ConstrainedFreeEnergyEntrySet":
        """

        Args:
            entry_set:
                Original FreeEnergyEntrySet.
            constr_chempot:
                Constrained chemical potentials in absolute scale.

        Returns:
            ConstrainedFreeEnergyEntrySet object.
        """
        new_entries = []
        for e in entry_set.entries:
            new_entry = deepcopy(e)
            comp = e.composition
            sub_comp = Composition({e: comp[e] for e in constr_chempot})
            if comp == sub_comp:
                continue
            new_entry.composition = comp - sub_comp
            sub_energy = sum([comp[e] * c for e, c in constr_chempot.items()])
            new_entry.energy = e.energy - sub_energy
            new_entries.append(new_entry)

        return cls(entries=new_entries,
                   original_entries=entry_set.entries,
                   constr_chempot=constr_chempot,
                   temperature=entry_set.temperature,
                   pressure=entry_set.pressure)





# -*- coding: utf-8 -*-

from pathlib import Path

from vise.util.testing import ViseTest
from vise.analyzer.band_plotter import (
    PrettyBSPlotter, ModBSPlotter, VaspBandStructureSymmLine)


class PlotBandStructureTest(ViseTest):
    def setUp(self) -> None:

        parent_dir = Path(__file__).parent

        self.mgo_band = VaspBandStructureSymmLine(
            parent_dir / "MgO_band_KPOINTS",
            parent_dir / "MgO_band_vasprun.xml", is_projection=True)
        self.mgo_nhse_band = \
            VaspBandStructureSymmLine(parent_dir / "MgO_band_KPOINTS",
                                      parent_dir / "MgO_band_nhse_vasprun.xml")

    def test(self):
        plotter = PrettyBSPlotter(band=self.mgo_band, band2=self.mgo_nhse_band,
                                  absolute=False)
        plotter.show()




# -*- coding: utf-8 -*-

import itertools
import os
import re
from copy import deepcopy
from pathlib import Path

from monty.io import zopen
from monty.serialization import loadfn

from pymatgen.electronic_structure.core import Magmom
from pymatgen.io.vasp import Incar
from pymatgen.util.io_utils import clean_lines

from tabulate import tabulate


MODULE_DIR = Path(os.path.dirname(os.path.abspath(__file__)))
# This incar_flags should be OrderedDict, but from python 3.6, dict uses
# order-preserving semantics. Furthermore, it does not affect vasp result.
incar_flags = loadfn(MODULE_DIR / "datasets" / "incar_flags.yaml")


class ViseIncar(Incar):
    """Incar class modified for pretty writing of INCAR file.

    Since from_file and from_string methods in Incar class use Incar class
    constructor, we need to override them.
    """

    @classmethod
    def from_file(cls, filename: str) -> "ViseIncar":
        """Reads an Incar object from a file.

        Args:
            filename (str): Filename for file

        Returns:
            ViseIncar object
        """
        with zopen(filename, "rt") as f:
            return cls.from_string(f.read())

    @classmethod
    def from_dict(cls, d: dict) -> "ViseIncar":
        kwargs = deepcopy(d)
        if kwargs.get("MAGMOM") and isinstance(kwargs["MAGMOM"][0], dict):
            kwargs["MAGMOM"] = [Magmom.from_dict(m) for m in kwargs["MAGMOM"]]

        return cls({k: v for k, v in d.items() if k not in ("@module",
                                                            "@class")})

    @classmethod
    def from_string(cls, string: str):
        """Reads an Incar object from a string.

        This method is different from that of Incar superclass at it does not
        support ";" semantic which split the incar flags.

        Args:
            string (str): Incar string

        Returns:
            ViseIncar object
        """
        lines = list(clean_lines(string.splitlines()))
        params = {}
        for line in lines:
            # YK: Support the split of ";" semantic in INCAR file
            for sub_line in line.split(";"):
                m = re.match(r'(\w+)\s*=\s*(.*)', sub_line)
                if m:
                    key = m.group(1).strip()
                    val = m.group(2).strip()
                    val = ViseIncar.proc_val(key, val)
                    params[key] = val

        return cls(params)

    def __add__(self, other: Incar) -> "ViseIncar":
        """Add all the values of another INCAR object to this object. """
        params = {k: v for k, v in self.items()}
        for k, v in other.items():
            if k in self and v != self[k]:
                raise ValueError("Incars have conflicting values!")
            else:
                params[k] = v
        return ViseIncar(params)

    def get_string(self, **kwargs) -> str:
        """Override for the pretty printing. """
        lines = []
        check_incar_keys = deepcopy(self)
        for key, val in incar_flags.items():
            comment = False
            blank_line = False
            ll = []

            for v in val:
                if v in check_incar_keys:
                    if comment is False:
                        lines.append(f"# {key} \n")
                        comment = True
                    if v == "MAGMOM" and isinstance(self[v], list):
                        value = []
                        if (isinstance(self[v][0], list) or
                            isinstance(self[v][0], Magmom)) and \
                                (self.get("LSORBIT") or
                                 self.get("LNONCOLLINEAR")):
                            value.append(
                                " ".join(str(i) for j in self[v] for i in j))
                        elif self.get("LSORBIT") or self.get("LNONCOLLINEAR"):
                            for m, g in itertools.groupby(self[v]):
                                value.append("3*{}*{}".format(len(tuple(g)), m))
                        # YK: Add this
                        elif len(self[v]) == 1:
                            value.append(str(self[v][0]))
                        else:
                            # float() to ensure backwards compatibility between
                            # float magmoms and Magmom objects
                            try:
                                for m, g in itertools.groupby(self[v],
                                                              lambda x: float(
                                                                  x)):
                                    value.append(
                                        "{}*{}".format(len(tuple(g)), m))
                            except ValueError:
                                raise ValueError("MAGMOM could be improper.")

                        ll.append([v, " ".join(value)])
                    elif isinstance(self[v], list):
                        ll.append([v, " ".join([str(i) for i in self[v]])])
                    else:
                        ll.append([v, self[v]])
                    blank_line = True
                    check_incar_keys.pop(v)
            if blank_line:
                # if not disable_numparse, LOPTICS = True becomes 1.
                lines.append(str(tabulate([[l[0], "=", l[1]] for l in ll],
                                          tablefmt="plain",
                                          disable_numparse=True)) + "\n")
                lines.append("\n")  # blank space

        for mson_key in ["@module", "@class"]:
            if mson_key in check_incar_keys:
                check_incar_keys.pop(mson_key)

        if check_incar_keys:
            raise ValueError(f"{check_incar_keys.keys()} are invalid in INCAR.")

        return "".join(lines)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import argparse
import sys
from typing import Union

from vise.cli.main_function import (
    get_poscar_from_mp, vasp_set, chempotdiag, plot_band, plot_dos, vasp_run,
    kpt_conv, band_gap)
from vise.config import SYMMETRY_TOLERANCE, ANGLE_TOL, KPT_DENSITY, TIMEOUT
from vise.custodian_extension.jobs import ViseVaspJob
from vise.input_set.input_set import ViseInputSet
from vise.util.logger import get_logger
from vise.cli.main_tools import dict2list, get_user_settings, get_default_args
from vise.util.mp_tools import make_poscars_from_mp
from vise.util.tools import str2bool
from vise import __version__

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"

logger = get_logger(__name__)

# The following keys are set by vise.yaml
setting_keys = ["vasp_cmd",
                "symprec",
                "angle_tolerance",
                "xc",
                "kpt_density",
                "initial_kpt_density",
                "vise_opts",
                "user_incar_settings",
                "ldauu",
                "ldaul",
                "potcar_set",
                "potcar_set_name",
                "max_relax_num",
                "removed_files",
                "left_files",
                "timeout"]


def simple_override(d: dict, keys: Union[list, str]) -> None:
    """Override dict if keys exist in vise.yaml.

    When the value in the user_settings is a dict, it will be changed to
    list using dict2list.
    """
    user_settings, yaml_path = get_user_settings(yaml_filename="vise.yaml",
                                                 setting_keys=setting_keys)
    if isinstance(keys, str):
        keys = [keys]

    for key in keys:
        if key in user_settings:
            v = user_settings[key]
            if isinstance(v, dict):
                v = dict2list(v)
            d[key] = v


def parse_args(args):

    parser = argparse.ArgumentParser(
        description="""                            
    Vise is a package that helps researchers to do first-principles calculations 
    with the VASP code.""",
        epilog=f"""                                 
    Author: Yu Kumagai \n
    Version: {__version__}""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    subparsers = parser.add_subparsers()

    # -- parent parser: prec
    prec = {"symprec": SYMMETRY_TOLERANCE,
            "angle_tolerance": ANGLE_TOL}
    simple_override(prec, ["symprec", "angle_tolerance"])

    prec_parser = argparse.ArgumentParser(description="Prec-related parser",
                                          add_help=False)
    prec_parser.add_argument(
        "--symprec", type=float, default=prec["symprec"],
        help="Set length precision used for symmetry analysis [A].")
    prec_parser.add_argument(
        "--angle_tolerance", type=float, default=prec["angle_tolerance"],
        help="Set angle precision used for symmetry analysis.")

    # -- parent parser:  custodian
    custodian_defaults = get_default_args(ViseVaspJob.kpt_converge)
    custodian_defaults["vasp_cmd"] = None
    custodian_defaults["timeout"] = TIMEOUT
    simple_override(custodian_defaults, ["vasp_cmd",
                                         "max_relax_num",
                                         "timeout",
                                         "left_files"])
    custodian_parser = argparse.ArgumentParser(
        description="Vasp custodian-related parser",
        add_help=False)

    custodian_parser.add_argument(
        "-v", "--vasp_cmd", nargs="+", type=str,
        default=custodian_defaults["vasp_cmd"],
        help="VASP command. If you are using mpirun, set this to something "
             "like \"mpirun pvasp\".",)
    custodian_parser.add_argument(
        "--handler_name", type=str, default="default",
        help="Custodian error handler name listed in error_handlers.")
    custodian_parser.add_argument(
        "--max_relax_num", default=custodian_defaults["max_relax_num"],
        type=int,
        help="Maximum number of relaxations.")
    custodian_parser.add_argument(
        "--left_files", type=str, nargs="+",
        default=custodian_defaults["left_files"],
        help="Filenames that are left at the calculation directory.")
    custodian_parser.add_argument(
        "--timeout", type=int, default=custodian_defaults["timeout"],
        help="Timeout used in TooLongTimeCalcErrorHandler.")
    custodian_parser.add_argument(
        "--remove_wavecar", action="store_true",
        help="Remove WAVECAR file after the calculation is finished.")

    # -- parent parser:  vasp set
    vasp_set_defaults = get_default_args(ViseInputSet.make_input)
    vasp_set_defaults.update(ViseInputSet.TASK_OPTIONS)
    vasp_set_defaults.update(ViseInputSet.XC_OPTIONS)
    vasp_set_defaults["potcar_set"] = None
    vasp_set_defaults["vise_opts"] = None
    vasp_set_defaults["user_incar_settings"] = None

    simple_override(vasp_set_defaults, ["xc",
                                        "task",
                                        "vise_opts",
                                        "user_incar_settings",
                                        "potcar_set",
                                        "potcar_set_name",
                                        "ldauu",
                                        "ldaul"])

    vasp_set_parser = argparse.ArgumentParser(
        description="Vasp set-related parser",
        add_help=False)

    vasp_set_parser.add_argument(
        "--potcar", dest="potcar_set", default=vasp_set_defaults["potcar_set"],
        type=str, nargs="+",
        help="User specifying POTCAR set. E.g., Mg_pv O_h")
    vasp_set_parser.add_argument(
        "--potcar_set_name", default=vasp_set_defaults["potcar_set_name"],
        type=str,
        help="User specifying POTCAR set, i ie. normal ,gw, or mp_relax_set.")
    vasp_set_parser.add_argument(
        "-x", "--xc", default=str(vasp_set_defaults["xc"]), type=str,
        help="Exchange-correlation (XC) interaction treatment.")
    vasp_set_parser.add_argument(
        "-t", "--task", default=str(vasp_set_defaults["task"]), type=str,
        help="The task name. See document of vise.")
    vasp_set_parser.add_argument(
        "--vise_opts", type=str, nargs="+",
        default=vasp_set_defaults["vise_opts"],
        help="Keyword arguments for options in make_input classmethod of "
             "ViseInputSet in vise. See document in vise for details.")
    vasp_set_parser.add_argument(
        "-uis", "--user_incar_settings", type=str, nargs="+",
        default=vasp_set_defaults["user_incar_settings"],
        help="user_incar_settings in make_input classmethod of ViseInputSet in "
             "vise. The default of this flag is set by the vise.yaml, so if one"
             "does not want to override the default, use "
             "additional_user_incar_setting instead. See also document in "
             "vise input_set for details.")
    vasp_set_parser.add_argument(
        "-auis", "--additional_user_incar_settings", type=str, nargs="+",
        default=None,
        help="Use this if one does not want to override user_incar_settings "
             "written in the yaml file")
    vasp_set_parser.add_argument(
        "--ldauu", type=str, default=vasp_set_defaults["ldauu"], nargs="+",
        help="Dict of LDAUU values")
    vasp_set_parser.add_argument(
        "--ldaul", type=str, default=vasp_set_defaults["ldaul"], nargs="+",
        help="Dict of LDAUL values.")
    vasp_set_parser.add_argument(
        "-c", "--charge", type=float, help="Charge state.")

    # -- get_poscars -----------------------------------------------------------
    parser_get_poscar = subparsers.add_parser(
        name="get_poscars",
        description="Tools for generating POSCAR file(s)",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['gp'])

    gp_defaults = get_default_args(make_poscars_from_mp)

    parser_get_poscar.add_argument(
        "-p", "--poscar", type=str,
        help="POSCAR-type file name.", metavar="FILE")
    parser_get_poscar.add_argument(
        "-n", "--number", type=int,
        help="MP entry number without prefix 'mp-'")
    parser_get_poscar.add_argument(
        "-e", "--elements", type=str, nargs="+",
        help="Create directories with POSCARs containing the input elements.")
    parser_get_poscar.add_argument(
        "--e_above_hull", type=float, default=gp_defaults["e_above_hull"],
        help="Collect materials with this hull energy in eV/atom.")
    parser_get_poscar.add_argument(
        "--molecules", type=str2bool, default=gp_defaults["molecules"],
        help="Whether one doesn't want to replace pmg structures to molecules.")

    parser_get_poscar.set_defaults(func=get_poscar_from_mp)

    # -- vasp_set ---------------------------------------------------------
    parser_vasp_set = subparsers.add_parser(
        name="vasp_set",
        parents=[vasp_set_parser, prec_parser],
        description="Tools for constructing vasp input set with vise",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['vs'])

    vs_defaults = {"kpt_density": KPT_DENSITY}
    simple_override(vs_defaults, ["kpt_density"])

    parser_vasp_set.add_argument(
        "--json", type=str,
        help="Json file name for printing the ViseInputSet info.")
    parser_vasp_set.add_argument(
        "-p", "--poscar", default="POSCAR", type=str,
        help="POSCAR-type file name.")
    parser_vasp_set.add_argument(
        "-k", "--kpt_density", default=vs_defaults["kpt_density"], type=float,
        help="K-point density in Angstrom along each direction .")
    parser_vasp_set.add_argument(
        "-s", "--standardize_structure", type=str2bool, default=True,
        help="Store if one doesn't want the cell to be transformed to a "
             "primitive cell.")
    parser_vasp_set.add_argument(
        "-pi", "--prior_info", type=str2bool, default=True,
        help="Whether to use prior_info.json when it exists.")
    parser_vasp_set.add_argument(
        "--dirs", nargs="+", type=str,
        help="Make vasp set for the directories in the same condition.")
    parser_vasp_set.add_argument(
        "-d", "--prev_dir", type=str,
        help="Inherit input files from the previous directory.")

    del vs_defaults

    parser_vasp_set.set_defaults(func=vasp_set)

    # -- vasp_run --------------------------------------------------------------
    parser_vasp_run = subparsers.add_parser(
        name="vasp_run",
        parents=[custodian_parser, prec_parser],
        description="Tools for vasp run",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['vr'])

    parser_vasp_run.add_argument(
        "-p", "--print", action="store_true",
        help="Whether to print str_opt.json.")
    parser_vasp_run.add_argument(
        "--json_file", default="str_opt.json", type=str,
        help="str_opt.json filename.")

    parser_vasp_run.set_defaults(func=vasp_run)

    # -- kpt_conv --------------------------------------------------------------
    parser_kpt_conv = subparsers.add_parser(
        name="kpt_conv",
        parents=[vasp_set_parser, custodian_parser, prec_parser],
        description="Tools for vasp run for k-point convergence",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['kc'])

    kc_defaults = get_default_args(ViseVaspJob.kpt_converge)
    kc_defaults["vasp_cmd"] = None

    simple_override(kc_defaults, ["vasp_cmd",
                                  "initial_kpt_density"
                                  "convergence_criterion"])

    parser_kpt_conv.add_argument(
        "-p", "--print", action="store_true",
        help="Whether to print str_opt.json.")
    parser_kpt_conv.add_argument(
        "--json_file", default="kpt_conv.json", type=str,
        help="kpt_conv.json filename.")
    parser_kpt_conv.add_argument(
        "-ikd", "--initial_kpt_density", type=float,
        default=kc_defaults["initial_kpt_density"],
        help="Initial k-point density in 1/A.")
    parser_kpt_conv.add_argument(
        "--criterion", dest="convergence_criterion",
        default=kc_defaults["convergence_criterion"], type=float,
        help="Convergence criterion of kpoints in eV/atom.")

    del kc_defaults
    parser_kpt_conv.set_defaults(func=kpt_conv)

    # -- chempotdiag -----------------------------------------------------------
    parser_cpd = subparsers.add_parser(
        name="chempotdiag",
        description="Tools for chemical potentials",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cpd'])

    # input
    # from file
    # parser_cpd.add_argument(
    #     "-e", "--energy", dest="energy_csv", type=str, default=None,
    #     help="Name of csv file of energies of compounds")
    parser_cpd.add_argument(
        "-dpd", "--draw_phase_diagram", action="store_true",
        help="set when drawing the phase diagram.")
    parser_cpd.add_argument(
        "-d", "--vasp_dirs", type=str, nargs='+',
        help="Drawing diagram from specified directories of vasp calculations")
    parser_cpd.add_argument(
        "-v", "--vasprun", default="vasprun.xml", type=str)
    parser_cpd.add_argument(
        "-e", "--elements", type=str, nargs="+",
        help="Element names. Obtain the total energies from the MP.")
    parser_cpd.add_argument(
        "-c", "--target_comp", type=str,
        help="Target compound focused for post process such as defect "
             "calculations.")
    parser_cpd.add_argument(
        "-f", dest="filename", type=str, help="Saved pdf file name.")
    parser_cpd.add_argument(
        "-pg", "--parse_gas", type=str2bool, default=True,
        help="Whether to parse results as gas for gas phases.")
    parser_cpd.add_argument(
        "-pp", "--partial_pressures", type=str, nargs='+',
        help="partial pressure of system in Pa. "
             "Example: -pp O2 1e+5 N2 20000 -> O2: 1e+5(Pa), N2: 20000(Pa)")
    parser_cpd.add_argument(
        "-t", "--temperature", type=float, default=0.0,
        help="temperature of system in K.")

    # thermodynamic status (P and T) input
    # # output
    # parser_cpd.add_argument("-y", "--yaml",
    #                     action="store_const", const=True,
    #                     default=False,
    #                     help="Dumps yaml of remarked_compound")

    parser_cpd.set_defaults(func=chempotdiag)

    # -- plot_band -----------------------------------------------------------
    parser_plot_band = subparsers.add_parser(
        name="plot_band",
        parents=[prec_parser],
        description="Tools for plotting band structures",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pb'])

    parser_plot_band.add_argument(
        "-v", dest="vasprun", default="vasprun.xml", type=str)
    parser_plot_band.add_argument(
        "-v2", dest="vasprun2", type=str)
    parser_plot_band.add_argument(
        "-k", dest="kpoints", default="KPOINTS", type=str)
    parser_plot_band.add_argument(
        "-y", dest="y_range", nargs="+", type=float,
        help="Energy range, requiring two values.")
    parser_plot_band.add_argument(
        "-f", dest="filename", type=str, default=None, help="pdf file name.")
    parser_plot_band.add_argument(
        "-a", dest="absolute", action="store_true",
        help="Show in the absolute energy scale.")
    parser_plot_band.add_argument(
        "-l", dest="legend", action="store_false",
        help="Not show the legend.")

    parser_plot_band.set_defaults(func=plot_band)

    # -- plot_dos -----------------------------------------------------------
    parser_plot_dos = subparsers.add_parser(
        name="plot_dos",
        parents=[prec_parser],
        description="Tools for plotting density of states",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pd'])

    parser_plot_dos.add_argument(
        "-v", dest="vasprun", type=str, default="vasprun.xml")
    parser_plot_dos.add_argument(
        "-cv", dest="cbm_vbm", type=float, nargs="+",
        help="Set CBM and VBM.")
    parser_plot_dos.add_argument(
        "-t", dest="pdos_type", type=str, default="element",
        help=".")
    parser_plot_dos.add_argument(
        "-s", dest="specific", type=str, nargs="+", default=None, help=".")
    parser_plot_dos.add_argument(
        "-o", dest="orbital", type=str2bool, default=True,
        help="Whether to decompose orbital components.")
    parser_plot_dos.add_argument(
        "-x", "--x_range", nargs="+", type=float, default=None,
        help="Set energy minimum and maximum.")
    parser_plot_dos.add_argument(
        "-y", "--ymaxs", nargs="+", type=float, default=None,
        help="Set max values of y ranges. Support two ways."
             "1st: total_max, all_the_atoms" 
             "2nd: total_max, 1st_atom, 2nd_atom, ...")
    parser_plot_dos.add_argument(
        "-f", dest="filename", type=str, default="dos.pdf",
        help="Pdf file name.")
    parser_plot_dos.add_argument(
        "-a", "--absolute", action="store_true",
        help="Set when showing the figure in the absolute energy scale.")
    parser_plot_dos.add_argument(
        "-l", "--legend", type=str2bool, default=True,
        help="Whether to show the figure legend.")
    parser_plot_dos.add_argument(
        "-c", "--crop_first_value", type=str2bool, default=True,
        help="Whether to crop the first value in DOS.")

    parser_plot_dos.set_defaults(func=plot_dos)

    # -- band_gap --------------------------------------------------------------
    parser_band_gap = subparsers.add_parser(
        name="band_gap",
        description="Calculate the band gap from vasprun.xml",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['bg'])

    parser_band_gap.add_argument(
        "-v", "--vasprun", type=str, default="vasprun.xml", metavar="FILE")
    parser_band_gap.add_argument(
        "-o", "--outcar", type=str, default="OUTCAR", metavar="FILE")
    parser_band_gap.set_defaults(func=band_gap)

    # try:
    #     import argcomplete
    #     argcomplete.autocomplete(parser)
    #     # This supports bash autocompletion. To enable this, pip install
    #     # argcomplete, activate global completion, or add
    #     #      eval "$(register-python-argcomplete vise)"
    #     # into your .bash_profile or .bashrc
    # except ImportError:
    #     pass

    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-

import json
import shutil
from pathlib import Path
from typing import List
import yaml

from pymatgen import Element, MPRester, Composition

from vise.util.logger import get_logger
from vise.chempotdiag.gas import Gas

logger = get_logger(__name__)


def get_mp_materials(elements: List[str],
                     properties: List[str],
                     e_above_hull: float = 1e-4,
                     api_key=None):
    """
    """
    exclude_z = list(i for i in range(1, 100))
    excluded_elements = [str(Element.from_Z(z)) for z in exclude_z]
    for e in elements:
        excluded_elements.remove(e)
    with MPRester(api_key) as m:
        materials = \
            m.query(criteria={"elements": {"$in": elements,
                                           "$nin": excluded_elements},
                              "e_above_hull": {"$lte": e_above_hull}},
                    properties=properties)

    return materials


def make_poscars_from_mp(elements,
                         path: Path = Path.cwd(),
                         e_above_hull=0.01,
                         api_key=None,
                         molecules=True,
                         only_molecules=True,
                         only_unary=False) -> None:
    """

    Args:
        elements(list): like ["Cu", "O"]
        path(str):
        e_above_hull(float):
        api_key(str):
        molecules(bool):
        only_molecules:
        only_unary:

    Returns:
        None
    """
    if not path.is_dir:
        raise NotADirectoryError(f"{path} is not directory.")

    mol_dir = Path(__file__).parent / ".." / "chempotdiag" / "molecules"

    molecules_formula_list = []
    if molecules:
        for g in Gas:
            comp = Composition(str(g))
            if set([str(e) for e in comp.elements]) < set(elements):
                molecules_formula_list.append(comp.reduced_formula)
                dirname = path / f"mol_{str(comp)}"
                if dirname.exists():
                    logger.critical(f"{dirname} exists! So, skip creating it.")
                else:
                    dirname.mkdir()
                    shutil.copyfile(mol_dir / str(comp) / "POSCAR",
                                    dirname / "POSCAR")
                    shutil.copyfile(mol_dir / str(comp) / "prior_info.yaml",
                                    dirname / "prior_info.yaml")

    properties = ["task_id",
                  "full_formula",
                  "final_energy",
                  "structure",
                  "spacegroup",
                  "band_gap",
                  "total_magnetization",
                  "magnetic_type"]
    materials = get_mp_materials(elements, properties, e_above_hull, api_key)

    for m in materials:
        comp = Composition(m.pop("full_formula"))
        formula = comp.reduced_formula
        if only_molecules and formula in molecules_formula_list:
            continue
        if only_unary and len(comp) == 1:
            continue

        m_path = path / f"{m['task_id']}_{formula}"
        m_path.mkdir()
        m.pop("structure").to(filename=m_path / "POSCAR")
        yaml_path = m_path / "prior_info.yaml"
        with open(str(yaml_path), "w") as fw:
            fw.write(yaml.dump(m))



#  Copyright (c) Oba-group
#  Distributed under the terms of the MIT License.

__copyright__    = 'Akira Takahashi'
__version__ = "v1.0.0"
__author__       = 'Akira Takahashi'
__author_email__ = 'chempotdiag.gmail'
__url__          = 'http://github.com/account/repository'



# -*- coding: utf-8 -*-
from copy import deepcopy

from pymatgen.io.vasp import Potcar

from vise.input_set.settings_potcar import XcTaskPotcar
from vise.input_set.xc import Xc

from vise.util.testing import ViseTest

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class XcTaskPotcarTest(ViseTest):
    def test_gga(self) -> None:
        xc_task_potcar = XcTaskPotcar.from_options(xc=Xc.pbe,
                                                   symbol_list=["Mg", "O"])

        expected = Potcar(["Mg", "O"], functional="PBE_54").as_dict()
        self.assertEqual(expected, xc_task_potcar.potcar.as_dict())
        self.assertEqual(400.0, xc_task_potcar.max_enmax)

    # def test_lda(self) -> None:
    #     xc_task_potcar = XcTaskPotcar.from_options(xc=Xc.lda,
    #                                                symbol_list=["Mg", "O"])

        # expected = Potcar(["Mg", "O"], functional="LDA").as_dict()
        # self.assertEqual(expected, xc_task_potcar.potcar.as_dict())

    def test_mp_relax_set(self) -> None:
        xc_task_potcar = \
            XcTaskPotcar.from_options(xc=Xc.pbe,
                                      symbol_list=["Mg", "O"],
                                      potcar_set_name="mp_relax_set")

        expected = Potcar(["Mg_pv", "O"], functional="PBE_54").as_dict()
        self.assertEqual(expected, xc_task_potcar.potcar.as_dict())

    def test_override(self) -> None:
        xc_task_potcar = \
            XcTaskPotcar.from_options(xc=Xc.pbe,
                                      symbol_list=["Mg", "O"],
                                      override_potcar_set={"Mg": "Mg_pv"})

        expected = Potcar(["Mg_pv", "O"], functional="PBE_54").as_dict()
        self.assertEqual(expected, xc_task_potcar.potcar.as_dict())



# -*- coding: utf-8 -*-

from custodian.vasp.interpreter import VaspModder

from vise.input_set.vasp_input import ViseVaspInput


class ViseVaspModder(VaspModder):
    def __init__(self, actions=None, strict=True, vi=None):
        """ Initializes a Modder for VaspInput sets

        Args:
            actions ([Action]):
            strict (bool)
            vi (vise.input_set.vasp_input.ViseVaspInput)
                See docstrings of constructor of VaspModder
        """
        if not vi:
            vi = ViseVaspInput.from_directory(".")
        super().__init__(actions, strict, vi)




# -*- coding: utf-8 -*-

from pymatgen.io.vasp import Potcar

from vise.input_set.settings_util import (
    load_potcar_yaml, load_default_incar_settings, check_keys, nelect, nbands)
from vise.input_set.settings_incar import (
    XC_REQUIRED_FLAGS, XC_OPTIONAL_FLAGS)
from vise.util.testing import ViseTest

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class LoadPotcarYamlTest(ViseTest):
    def setUp(self):
        self.normal = load_potcar_yaml("normal")
        self.normal_override = load_potcar_yaml(
            set_name="normal", override_potcar_set={"Zr": "Zr"})

    def test(self):
        self.assertEqual("Zr_sv", self.normal["Zr"])

    def test_override(self):
        self.assertEqual("Zr", self.normal_override["Zr"])

    def test_raise_error(self):
        with self.assertRaises(KeyError):
            load_potcar_yaml("not_exist_set_name")


class LoadDefaultIncarSettingsTest(ViseTest):
    def setUp(self):
        self.hse = \
            load_default_incar_settings(yaml_filename="xc_incar_set.yaml",
                                        required_flags=XC_REQUIRED_FLAGS,
                                        optional_flags=XC_OPTIONAL_FLAGS,
                                        key_name="hse")

    def test(self):
        expected = {'LWAVE': True,
                    'ALGO': 'D',
                    'PRECFOCK': 'Fast',
                    'HFSCREEN': 0.208,
                    'LHFCALC': True,
                    'TIME': 0.4}
        self.assertEqual(expected, self.hse)


class CheckKeysTest(ViseTest):
    def test_succeed(self):
        d = {"a": 1, "b": 2, "c": 3}
        self.assertTrue(check_keys(d=d, required={"a"}, optional={"b", "c"}))

    def test_raise_error1(self):
        d = {"a": 1, "b": 2, "c": 3}
        with self.assertRaises(KeyError):
            check_keys(d=d, required={"a", "d"}, optional={"b", "c"})

    def test_raise_error2(self):
        d = {"a": 1, "b": 2, "c": 3}
        with self.assertRaises(KeyError):
            check_keys(d=d, required={"a"}, optional={"b"})


class NelectTest(ViseTest):
    def setUp(self):
        self.mgo = self.get_structure_by_name("MgO").composition
        self.potcar = Potcar(["Mg", "O"], functional="PBE_54")

    def test_neutral(self):
        expected = 8
        self.assertEqual(expected, nelect(self.mgo, self.potcar))

    def test_charged(self):
        expected = 9
        self.assertEqual(expected, nelect(self.mgo, self.potcar, charge=-1))


class NbandsTest(ViseTest):
    def setUp(self):
        self.mgo = self.get_structure_by_name("MgO").composition
        self.potcar = Potcar(["Mg", "O"], functional="PBE_54")

    def test_neutral(self):
        expected = 8 / 2 + 4 + 4
        self.assertEqual(expected, nbands(self.mgo, self.potcar))




# -*- coding: utf-8 -*-

from vise.util.testing import ViseTest
from vise.input_set.settings_structure_kpoints import TaskStructureKpoints
from vise.input_set.task import Task


class TaskStructureKpointsTest(ViseTest):
    def setUp(self) -> None:
        mgo = self.get_structure_by_name("MgO")
        self.structure_opt = \
            TaskStructureKpoints.from_options(task=Task.band,
                                              original_structure=mgo,
                                              standardize_structure=True,
                                              sort_structure=True,
                                              is_magnetization=False,
                                              kpt_mode="primitive",
                                              kpt_density=3,
                                              kpt_shift=[0, 0, 0],
                                              only_even=True,
                                              band_ref_dist=0.025,
                                              factor=1,
                                              symprec=0.01,
                                              angle_tolerance=3)

    def test(self):
        expected = 287
        actual = self.structure_opt.kpoints.num_kpts
        self.assertEqual(expected, actual)

        expected = ['GAMMA', 'X', 'U', 'K', 'GAMMA', 'L', 'W', 'X']
        actual = [l for l in self.structure_opt.kpoints.labels if l]
        self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import json
import string
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
import numpy as np
from typing import Optional, List, Dict

from monty.json import MSONable, MontyEncoder
from monty.serialization import loadfn

from pymatgen.analysis.phase_diagram import PhaseDiagram, PDEntry
from pymatgen.core.composition import Composition
from pymatgen.core.periodic_table import Element
from pymatgen.util.string import latexify

from vise.util.logger import get_logger

logger = get_logger(__name__)


class ChemPotDiag(MSONable):
    """chemical potential diagrams"""
    def __init__(self,
                 elements: List[Element],
                 el_ref_list: List[PDEntry],
                 dim: int,
                 vertices: List[List[float]],
                 qhull_entries: List[PDEntry],
                 comp_facets: List[List[int]],
                 target_comp: str,
                 target_comp_chempot: Dict[str, List[float]],
                 unstable_energy: Optional[float] = None):
        """

        Note: Chemical potential is in eV/*atom* even for the complex diagrams.

        Args:
            elements (List[Element]):
                List of pmg Element involved.
            el_ref_list (List[PDEntry]):
                List of PDEntry for each Element.
            dim (int):
                Dimension of chemical potential diagram, or number of elements.
            vertices (List[Dict[Element, float]]):
                Vertices in the CPD surrounded by compounds.
            qhull_entries (List[PDEntry]):
                Entries comprising convex hull.
            comp_facets (List[List[int]]):
               List of vertex indices comprising each composition.
               Indices are the same as qhull_entries, therefore composition
               and name are obtained via qhull_entries.
            target_comp (Composition):
                Target Composition.
            target_comp_chempot (Dict[str, Dict[Element, float]]):
                Chemical potentials at the vertices of the target composition.
                The keys are alphabets.
        """
        self.elements = elements
        self.el_ref_list = el_ref_list
        self.dim = dim
        self.vertices = vertices
        self.qhull_entries = qhull_entries
        self.comp_facets = comp_facets
        self.target_comp = target_comp
        self.target_comp_chempot = target_comp_chempot
        self.unstable_energy = unstable_energy

    @classmethod
    def from_phase_diagram(
            cls,
            pd: PhaseDiagram,
            target_comp: Optional[str] = None,
            allow_unstable_target_chempot: bool = False,
            ) -> "ChemPotDiag":
        """Create class instance object from pymatgen PhaseDiagram object

        Args:
            pd (PhaseDiagram):
                Pmg PhaseDiagram object.
            target_comp (None/str):
                Target composition.
            allow_unstable_target_chempot:
                Whether to allow unstable target chemical potential.
                The vertex is defined at the point where the surrounding
                competing phases in the phase diagram equilibrates.
        """
        elements = sorted(pd.elements)
        vertices = []
        adjoin_comps = []  # List of adjoining compositions to each vertex.

        for fc in pd.facets:
            comp_list = [pd.qhull_entries[i].composition for i in fc]
            e_list = [pd.get_form_energy_per_atom(pd.qhull_entries[i])
                      for i in fc]
            atomic_frac = [[c.get_atomic_fraction(e) for e in elements]
                           for c in comp_list]
            chempot = dict(zip(elements, np.linalg.solve(atomic_frac, e_list)))

            vertices.append([round(chempot[e], 10) for e in elements])
            adjoin_comps.append([pd.qhull_entries[i].composition for i in fc])

        comp_facets = []  # List of vertex indices comprising each composition.
        for e in pd.qhull_entries:
            comp_facets.append(
                [i for i, cs in enumerate(adjoin_comps) if e.composition in cs])

        target_comp_chempot = {}
        unstable_energy = None
        if target_comp:
            target_inds = [i for i, e in enumerate(pd.qhull_entries)
                           if Composition(e.name) == Composition(target_comp)]

            alphabet_list = list(string.ascii_uppercase)
            if not target_inds:
                if not allow_unstable_target_chempot:
                    raise ValueError(f"Target composition {target_comp} "
                                     f"is invalid. Choose from "
                                     f"{[e.name for e in pd.qhull_entries]}")
                else:
                    target_inds = \
                        [e for e in pd.unstable_entries
                         if Composition(e.name) == Composition(target_comp)]
                    if not target_inds:
                        raise ValueError(
                            f"Target composition {target_comp} does not "
                            f"exist even in unstable entries. Choose from "
                            f"{[e.name for e in pd.unstable_entries]}")
                    comp = target_inds[0].composition
                    unstable_energy = pd.get_hull_energy(comp)
                    # Find the facet in phase diagram where comp belongs to.
                    c = pd.pd_coords(comp)
                    facets = []
                    for f, s in zip(pd.facets, pd.simplexes):
                        if s.in_simplex(c, PhaseDiagram.numerical_tol / 10):
                            facets.append(f)

                    # Find the index of vertices in chemical potential diagram.
                    for i, fc in enumerate(facets):
                        ind = [v.tolist() for v in pd.facets].index(fc.tolist())
                        target_comp_chempot[alphabet_list[i]] = vertices[ind]
                        logger.warning(
                            f"Unstable compound {comp} is evaluated in "
                            f"ChemPotDiag. Unstable energy: {unstable_energy}.")
            else:
                if len(comp_facets) > 26:
                    raise ValueError(f"Too many vertices: {len(comp_facets)}.")

                for i, fc in enumerate(comp_facets[target_inds[0]]):
                    target_comp_chempot[alphabet_list[i]] = vertices[fc]

        return cls(elements=elements,
                   el_ref_list=[pd.el_refs[e] for e in elements],
                   dim=len(elements),
                   vertices=vertices,
                   qhull_entries=pd.qhull_entries,
                   comp_facets=comp_facets,
                   target_comp=target_comp,
                   target_comp_chempot=target_comp_chempot,
                   unstable_energy=unstable_energy)

    @property
    def target_comp_abs_chempot(self) -> Dict[str, List[float]]:
        """Same as target_comp_chempot but in absolute scale.

        Mostly used for calculating chemical potential-related properties such
        as point defects.
        """
        abs_chempot = {}
        for point, chempot in self.target_comp_chempot.items():
            abs_chempot[point] = [round(e.energy_per_atom + c, 10)
                                  for e, c in zip(self.el_ref_list, chempot)]
        return abs_chempot

    def draw_diagram(self,
                     title: str = None,
                     filename: Optional[str] = None) -> plt:
        """Draw chemical potential diagram.

        Args:
            title (str):
                Title of the diagram.
            filename (None/str):
                Specify this when saving the diagram as an image file.

        Returns:
            Pyplot
        """
        draw_range = min(-1.0, min(sum([i for i in self.vertices], [])) * 1.1)

        if self.dim == 2 or self.dim == 3:
            ax = self._plot(draw_range, title)
        else:
            raise NotImplementedError(f"Drawing {self.dim} is impossible.")

        if self.dim == 2:
            ax.set_xlim(draw_range, 0)
            ax.set_ylim(draw_range, 0)
            plt.gca().set_aspect('equal', adjustable='box')
        else:
            ax.set_xlim3d(draw_range, 0)
            ax.set_ylim3d(0, draw_range)
            ax.set_zlim3d(draw_range, 0)

        ax.grid(color='b', alpha=0.2, linestyle='dashed', linewidth=0.5)
        plt.tight_layout()

        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def _plot(self, draw_range: float, title: str):

        plt.rcParams['pdf.fonttype'] = 42
        plt.rcParams['ps.fonttype'] = 42
        plt.rcParams['svg.fonttype'] = "none"

        if self.dim == 2:
            ax = plt.figure().add_subplot(111)
        else:
            ax = plt.figure().add_subplot(111, projection='3d')

        text_kwargs = {}
        species = ["NA"] * self.dim
        for i, vertex_inds in enumerate(self.comp_facets):
            name = self.qhull_entries[i].name
            label = latexify(name)
            coords = [self.vertices[i] for i in vertex_inds]
            if len(coords) == 0:
                continue
            coords = np.array(coords).transpose()
            comp = self.qhull_entries[i].composition

            comp_color = "black"
            if self.target_comp and name == self.target_comp:
                comp_color = "blue"

            if self.dim == 2:
                if len(comp.elements) == 1:
                    ind = 0 if comp.elements[0] == self.elements[0] else 1
                    # [0, draw_range] & [draw_range, 0] for 1st & 2nd elements.
                    coord = [[draw_range * ind], [draw_range * (1 - ind)]]
                    coords = np.append(coords, coord, axis=1)
                    set_label = f"set_{['x', 'y'][ind]}label"
                    getattr(ax, set_label)(f"Chemical potential of {label}")
                    species[ind] = label

                plt.plot(*coords, zorder=1, color=comp_color, linewidth=3)

            else:  # 3-dim
                if len(comp.elements) == 1:
                    ex_inds = []
                    for ind, e in enumerate(self.elements):
                        if e in comp.elements:
                            label_ind = ind
                            set_label = f"set_{['x', 'y', 'z'][ind]}label"
                        else:
                            ex_inds.append(ind)

                    # Project vertices to the two edges and add one corner
                    c1 = coords.copy()
                    c2 = coords.copy()
                    c1[ex_inds[0], :] = c2[ex_inds[1], :] = draw_range
                    c3 = np.array([[0.0], [0.0], [0.0]])
                    c3[ex_inds[0], 0] = c3[ex_inds[1], 0] = draw_range
                    coords = np.concatenate((coords, c1, c2, c3), axis=1)
                    getattr(ax, set_label)(f"Chemical potential of {label}")
                    species[label_ind] = label

                elif len(comp.elements) == 2:
                    for ind, e in enumerate(self.elements):
                        if not any([str(e) == str(e2) for e2 in comp.elements]):
                            break
                    c = coords.copy()
                    c[ind, :] = draw_range
                    coords = np.append(coords, c, axis=1)

                sorted_t_coords = sort_coords(np.transpose(coords))
                face = Poly3DCollection([sorted_t_coords])

                color = [(comp.get_atomic_fraction(e) + 3) / 4
                         for e in self.elements]
                face.set_color(color)
                face.set_edgecolor("black")
                ax.add_collection3d(face)
                text_kwargs.update({"ha": "center", "va": "center"})

            center = np.average(coords, axis=1)
            ax.text(*center, label, size='smaller', zorder=100,
                    color=comp_color, **text_kwargs)

        title = title or f"Chemical potential diagram of {'-'.join(species)}"
        ax.set_title(title)

        for label, cp in self.target_comp_chempot.items():
            ax.text(*cp, label, zorder=200, color="blue", weight="bold")

        return ax

    @classmethod
    def load_json(cls, filename: str = "cpd.json"):
        return loadfn(filename)

    def to_json_file(self, filename: str = "cpd.json"):
        with open(filename, 'w') as fw:
            json.dump(self.as_dict(), fw, indent=2, cls=MontyEncoder)


def sort_coords(coords: np.ndarray) -> np.ndarray:
    """Sort coordinates based on the angle with first coord from the center.

    Args:
        coords (np.ndarray):
            Coordinates to be sorted. The format of coords is as follows.
            np.array([[x1, y1, z1], [x2, y2, z2], [x3, y3, z3]]

    Returns:
        np.ndarray for sorted coordinates.
    """
    if len(coords[0]) != 3:
        raise ValueError("Only valid for 3D vector")

    center = np.average(coords, axis=0)
    rel_coords = coords - center
    n = np.cross(rel_coords[0], rel_coords[1])
    if abs(np.linalg.norm(n)) < 1e-8:  # Skip parallel vectors.
        n = np.cross(rel_coords[0], rel_coords[2])
    n = n / np.linalg.norm(n)

    def angle_between_v0(index: int) -> float:
        """
        Args:
            index (int): index of coords.

        Returns (float):
            Angle between rays from the center to rel_coords[0] and
            rel_coords[int].
        """
        v0 = rel_coords[0] / np.linalg.norm(rel_coords[0])
        v = rel_coords[index] / np.linalg.norm(rel_coords[index])
        det = np.linalg.det(np.concatenate(([v0], [v], [n]), axis=0))
        angle = np.arctan2(np.clip(np.dot(v0, v), -1.0, 1.0), det)
        return angle

    indices = [i for i in range(len(coords))]
    indices.sort(key=angle_between_v0)

    return coords[indices]



# -*- coding: utf-8 -*-

import shutil

from vise.util.mp_tools import make_poscars_from_mp
from vise.util.testing import ViseTest


class TestGetMpMaterials(ViseTest):
    pass


class TestMakePoscarsFromMp(ViseTest):
    def test(self):
        make_poscars_from_mp(elements=["Mg", "O"])

    # uncomment these if one wants to check the created directories.
    def tearDown(self) -> None:
        shutil.rmtree("mol_O2")
        shutil.rmtree("mp-1265_MgO")
        shutil.rmtree("mp-1094122_Mg")


# -*- coding: utf-8 -*-

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class InvalidFileError(Exception):
    """Raised when the given file is invalid."""
    pass


class InvalidStructureError(Exception):
    """Raised when the structure is invalid."""
    pass


class NoVaspCommandError(Exception):
    """Raised when the vasp command is not set."""
    pass


class VaspNotConvergedError(Exception):
    """Raised when the vasp calculation is not converged."""
    pass


class KptNotConvergedError(Exception):
    """Raised when the k-point set is not converged."""
    pass



# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Optional

import numpy as np
from pymatgen.core.structure import Structure
from pymatgen.io.vasp import Kpoints
from vise.input_set.make_kpoints import MakeKpoints
from vise.input_set.task import Task, SPECTRA_TASK
from vise.util.logger import get_logger
from vise.util.structure_handler import find_spglib_primitive, get_symbol_list

logger = get_logger(__name__)
SET_DIR = Path(__file__).parent / "datasets"


class TaskStructureKpoints:

    def __init__(self,
                 structure: Structure,
                 kpoints: Kpoints,
                 is_structure_changed: bool,
                 sg: int,
                 num_kpts: int,
                 factor: int):
        self.structure = structure
        self.kpoints = kpoints
        self.is_structure_changed = is_structure_changed
        self.sg = sg
        self.num_kpts = num_kpts
        self.factor = factor

    @classmethod
    def from_options(cls,
                     task: Task,
                     original_structure: Structure,
                     standardize_structure: bool,
                     sort_structure: bool,
                     is_magnetization: bool,
                     kpt_mode: str,
                     kpt_density: float,
                     kpt_shift: list,
                     only_even: bool,
                     band_ref_dist: float,
                     factor: Optional[int],
                     symprec: float,
                     angle_tolerance: float):
        """Construct Structure and Kpoints from task and some options.

        When task and kpt_mode are not consistent e.g., task=Task.band,
        kpt_mode="manual", task is prioritized.

        Args: See ViseInputSet docstrings

        Returns:
            TaskStructureKpoints instance object
        """
        if sort_structure:
            structure = original_structure.get_sorted_structure()
            symbol_list = get_symbol_list(structure)
            orig_symbol_list = get_symbol_list(original_structure)
            if symbol_list != orig_symbol_list:
                logger.warning(
                    "CAUTION: The sequence of the species is changed. \n"
                    f"Symbol set in the original structure {symbol_list} \n"
                    f"Symbol set in the generated structure {orig_symbol_list}")
        else:
            structure = original_structure.copy()

        is_structure_changed = False
        manual_kpts = None
        if task == Task.defect:
            kpt_mode = "manual_set"
        elif task == Task.cluster_opt:
            kpt_mode = "manual_set"
            manual_kpts = [1, 1, 1]
            only_even = False
            kpt_shift = [0, 0, 0]
        elif task == Task.band:
            kpt_mode = "band"
        elif task == Task.phonon_force:
            kpt_mode = "manual_set"
            if kpt_shift != [0, 0, 0]:
                logger.warning("For phonon force calculations, Gamma centering "
                               "is forced for k-point sampling.")
            kpt_shift = [0, 0, 0]
        else:
            primitive_structure, is_structure_changed = \
                find_spglib_primitive(structure, symprec, angle_tolerance)

            if standardize_structure:
                org = original_structure.lattice.matrix
                primitive = primitive_structure.lattice.matrix
                if is_structure_changed:
                    with np.printoptions(precision=3, suppress=True):
                        logger.warning(
                            "CAUTION: The structure is changed.\n"
                            f"Original lattice\n {org} \n"
                            f"Generated lattice\n {primitive} \n")
                structure = primitive_structure
            else:
                if is_structure_changed and kpt_mode != "manual_set":
                    logger.warning(
                        "Standardizaion is set to False and the given "
                        "structure is not a primitive cell. Thus, the "
                        "kpoint set is switched to manual_set.")
                    kpt_mode = "manual_set"
                else:
                    logger.info("The structure is a standardized primitive.")

        # Gamma-center mesh is a must for GW calculations due to vasp
        # implementation and tetrahedron method, while is a strong recommend for
        # dos and dielectric function to sample the band edges.
        if task in SPECTRA_TASK and kpt_shift != [0, 0, 0]:
            logger.warning("Gamma centering is forced for k-point sampling.")
            kpt_shift = [0, 0, 0]

        if factor is None:
            if task == Task.dielectric_function:
                factor = 3
            elif task in (Task.dos, Task.dielectric_dfpt,
                          Task.dielectric_finite_field):
                factor = 2
            else:
                factor = 1

        kpoints = MakeKpoints(mode=kpt_mode,
                              structure=structure,
                              kpt_density=kpt_density,
                              only_even=only_even,
                              manual_kpts=manual_kpts,
                              ref_distance=band_ref_dist,
                              kpt_shift=kpt_shift,
                              factor=factor,
                              symprec=symprec,
                              angle_tolerance=angle_tolerance,
                              is_magnetization=is_magnetization)
        kpoints.make_kpoints()

        return cls(structure=structure,
                   kpoints=kpoints.kpoints,
                   is_structure_changed=kpoints.is_structure_changed,
                   sg=kpoints.sg,
                   num_kpts=kpoints.num_kpts,
                   factor=factor)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pathlib import Path
import tempfile

from vise.cli.main_tools import (
    potcar_str2dict, list2dict, dict2list, get_user_settings, get_default_args)
from vise.util.testing import ViseTest


class PotcarStr2DictTest(ViseTest):
    def test_none(self):
        self.assertEqual({}, potcar_str2dict(None))

    def test_list(self):
        expected = {"Mg": "Mg_pv", "O": "O_h"}
        actual = potcar_str2dict(["Mg_pv", "O_h"])
        self.assertEqual(expected, actual)

    def test_str(self):
        expected = {"Mg": "Mg_pv"}
        actual = potcar_str2dict("Mg_pv")
        self.assertEqual(expected, actual)

    def test_mutliple_potcars_for_same_element_error(self):
        with self.assertRaises(ValueError):
            potcar_str2dict(["Mg_pv", "Mg"])

    def test_incorrect_element_potcar_error(self):
        with self.assertRaises(ValueError):
            potcar_str2dict(["MgHH", "Mg"])


class List2DictTest(ViseTest):
    def setUp(self) -> None:
        self.key_candidates = ["ENCUT", "MAGMOM", "LWAVE"]

    def test_dict2list(self):
        flattened_list = ["ENCUT", "500", "MAGMOM", "4", "4.0", "LWAVE", "F"]
        actual = list2dict(flattened_list, self.key_candidates)
        expected = {"ENCUT": 500, "MAGMOM": [4, 4.0], "LWAVE": False}
        self.assertEqual(expected, actual)

    def test_fail(self):
        flattened_list = ["ENCAT", "500"]
        with self.assertRaises(ValueError):
            list2dict(flattened_list, self.key_candidates)

    def test_fail2(self):
        flattened_list = ["ENCUT", "500", "MAGMOM"]
        with self.assertRaises(ValueError):
            list2dict(flattened_list, self.key_candidates)


class Dict2ListTest(ViseTest):
    def test_dict2list(self):
        actual = dict2list({"a": 1, "b": "2 3 4", "c": True})
        expected = ["a", "1", "b", "2", "3", "4", "c", "True"]
        self.assertEqual(expected, actual)


class GetUserSettingsTest(ViseTest):
    def setUp(self) -> None:
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as tmp_file:
            tmp_file.write("""# This file is needed for the unittest.
symprec: 0.01
user_incar_settings:
  ALGO: False
ldauu:
  Mn: 5
  O: 3
xc: hse
potcar_set: Mg_pv O_h
dir: ./path""")
            # This is needed to actually write down the string.
            tmp_file.flush()
            self.setting_keys = ["symprec", "user_incar_settings", "ldauu",
                                 "xc", "potcar_set", "test_key", "dir"]
            self.user_setting, path = \
                get_user_settings(tmp_file.name, self.setting_keys)

    def test_dict(self):
        actual = self.user_setting['ldauu']
        expected = {'Mn': 5, 'O': 3}
        self.assertEqual(expected, actual)

    def test_bool(self):
        actual = self.user_setting['user_incar_settings']['ALGO']
        self.assertEqual(False, actual)

    def test_float(self):
        actual = self.user_setting['symprec']
        self.assertEqual(0.01, actual)

    def test_string(self):
        actual = self.user_setting['xc']
        self.assertEqual("hse", actual)

    def test_string_list(self):
        actual = self.user_setting['potcar_set']
        self.assertEqual("Mg_pv O_h", actual)

    def test_full_path(self):
        actual = self.user_setting['dir']
        cwd = Path.cwd()
        expected = cwd / "path"
        self.assertEqual(str(expected), actual)

    def test_fail(self):
        with self.assertRaises(ValueError):
            with tempfile.NamedTemporaryFile(mode="w",
                                             suffix=".yaml") as tmp_file:
                tmp_file.write("incorrect_key: 0.01")
                # This is needed to actually write down the string.
                tmp_file.flush()
                get_user_settings(tmp_file.name, self.setting_keys)


class GetDefaultArgsTest(ViseTest):
    def setUp(self) -> None:
        def test_func(a, b=1, c=True):
            pass

        class test_cls:
            def __init__(self, d, e=2, f=None):
                pass
            def test_method(self, g, h=3.0):
                pass

        self.default_func = get_default_args(test_func)
        self.default_cls = get_default_args(test_cls)
        self.default_method = get_default_args(test_cls.test_method)

    def test_func(self):
        expected = {"b": 1, "c": True}
        self.assertEqual(expected, self.default_func)

    def test_cls(self):
        expected = {"e": 2, "f": None}
        self.assertEqual(expected, self.default_cls)

    def test_method(self):
        expected = {"h": 3.0}
        self.assertEqual(expected, self.default_method)


# -*- coding: utf-8 -*-

from vise.util.testing import ViseTest
from vise.input_set.incar import incar_flags

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class IncarFlagsTest(ViseTest):

    def test(self) -> None:
        expected = ["SMASS", "NBLOCK"]
        actual = incar_flags["md"]
        self.assertEqual(expected, actual)





# -*- coding: utf-8 -*-

from distutils.util import strtobool
from typing import Callable, Any
from xml.etree.ElementTree import ParseError

from vise.util.logger import get_logger

logger = get_logger(__name__)


def parse_file(class_method_name: Callable, parsed_filename: str) -> Any:
    """Check filename and parse and return cls via __init__ or class method.

    Args:
         class_method_name (Callable):
            Method to parse the given file. E.g., CLASS.from_file
        parsed_filename (str):
            Parsed file name.

    Return:
         Return of the class method.
    """
    try:
        logger.info(f"Parsing {parsed_filename}...")
        return class_method_name(parsed_filename)
    except ParseError:
        logger.warning(f"Parsing {parsed_filename} failed.")
        raise ParseError
    except FileNotFoundError:
        logger.warning(f"File {parsed_filename} not found.")
        raise


def str2bool(string: str) -> bool:
    return bool(strtobool(string))


def is_str_digit(n: str) -> bool:
    """Check whether the given string is a digit or not.

    Args:
        n (str): The checked string.

    Returns:
        Bool.
    """
    try:
        float(n)
        return True
    except ValueError:
        return False


def is_str_int(n: str) -> bool:
    """Check whether the given string is an integer or not.

    Args:
        n (str): The checked string.

    Returns:
        Bool.
    """
    try:
        if int(n) - float(n) < 1e-5:
            return True
        else:
            return False
    except ValueError:
        return False


# -*- coding: utf-8 -*-

import os

from vise.input_set.input_set import ViseInputSet
from vise.input_set.xc import Xc
from vise.util.testing import ViseTest

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class InputSetTest(ViseTest):
    def setUp(self) -> None:
        mgo = self.get_structure_by_name("MgO")
        self.input_set = ViseInputSet.make_input(structure=mgo, xc=Xc.hse)

    def test_write(self):
        self.input_set.write_input(output_dir=".")
        os.remove("INCAR")
        os.remove("POSCAR")
        os.remove("POTCAR")
        os.remove("KPOINTS")
        os.remove("vise.json")

    def test_dict(self):
        expected = self.input_set.as_dict()
        actual = ViseInputSet.from_dict(expected).as_dict()
        self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-

from enum import unique, Enum


@unique
class Task(Enum):
    """ Supported tasks """
    structure_opt = "structure_opt"
    structure_opt_rough = "structure_opt_rough"
    structure_opt_tight = "structure_opt_tight"
    cluster_opt = "cluster_opt"
    phonon_force = "phonon_force"
    defect = "defect"
    band = "band"
    dos = "dos"
    dielectric_dfpt = "dielectric_dfpt"
    dielectric_finite_field = "dielectric_finite_field"
    dielectric_function = "dielectric_function"

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s):
        for m in Task:
            if m.value == s:
                return m
        raise AttributeError(f"Task: {s} is not proper.\n"
                             f"Supported Task:\n {cls.name_list()}")

    @classmethod
    def name_list(cls):
        return ', '.join([e.value for e in cls])


LATTICE_RELAX_TASK = (Task.structure_opt, Task.structure_opt_rough,
                      Task.structure_opt_tight)
PLOT_TASK = (Task.band, Task.dos, Task.dielectric_function)
SPECTRA_TASK = (Task.dos, Task.dielectric_function)




# -*- coding: utf-8 -*-

import unittest
import numpy as np
from unittest.mock import patch

from pymatgen.core.composition import Composition
from pymatgen.core.sites import Element
from pymatgen.analysis.phase_diagram import (
    PDEntry, PhaseDiagram, CompoundPhaseDiagram, PDPlotter)

from vise.chempotdiag.chem_pot_diag import ChemPotDiag, sort_coords
from vise.util.testing import ViseTest


DISABLE_DISPLAY_DIAGRAM = False


class TestChemPotDiag(ViseTest):

    def setUp(self) -> None:

        mg = PDEntry(Composition("Mg"), -1.0)
        mg2 = PDEntry(Composition("Mg"), -0.5)
        ca = PDEntry(Composition("Ca"), -2.0)
        sr = PDEntry(Composition("Sr"), -3.0)
        o = PDEntry(Composition("O"), -4.0)

        camg = PDEntry(Composition("Ca2Mg2"), -16.0)  # rel -10.0
        camgo = PDEntry(Composition("CaMgO"), -17.0)  # rel -10.0
        camgo2 = PDEntry(Composition("CaMgO2"), -11.0)  # rel 0.0
        camgsro4 = PDEntry(Composition("CaMgSrO4"), -32.0)  # rel -10

        self.pd_1d = PhaseDiagram([mg, mg2])
        self.pd_2d = PhaseDiagram([mg, ca, camg])
        self.pd_3d = PhaseDiagram([mg, ca, camg, o, camgo, camgo2])
        self.pd_4d = PhaseDiagram([mg, ca, camg, o, camgo, sr, camgsro4])

        self.cpd_1d = ChemPotDiag.from_phase_diagram(pd=self.pd_1d,
                                                     target_comp="Mg")
        self.cpd_2d = ChemPotDiag.from_phase_diagram(pd=self.pd_2d,
                                                     target_comp="CaMg")
        self.cpd_3d = ChemPotDiag.from_phase_diagram(pd=self.pd_3d,
                                                     target_comp="CaMgO")
        self.cpd_3d_unstable = \
            ChemPotDiag.from_phase_diagram(pd=self.pd_3d,
                                           target_comp="CaMgO2",
                                           allow_unstable_target_chempot=True)
        self.cpd_4d = ChemPotDiag.from_phase_diagram(pd=self.pd_4d,
                                                     target_comp="SrCaMgO4")
#        target_comp="CaMgSrO4")

        self.comp_pd = CompoundPhaseDiagram(
            entries=[mg, ca, o, camg, camgo, camgo2],
            terminal_compositions=[Composition("CaMg"),
                                   Composition("Ca"),
                                   Composition("O")])

        self.comp_cpd = ChemPotDiag.from_phase_diagram(
            pd=self.comp_pd,
            target_comp="MgCaO2",
            allow_unstable_target_chempot=True)

    def test_cpd_1d(self):
        self.assertEqual([Element.Mg], self.cpd_1d.elements)
        self.assertEqual(1, self.cpd_1d.dim)
        self.assertEqual([[0.0]], self.cpd_1d.vertices)
        self.assertEqual("Mg", self.cpd_1d.target_comp)
        self.assertEqual({'A': [0.0]}, self.cpd_1d.target_comp_chempot)
        self.assertEqual({'A': [-1.0]}, self.cpd_1d.target_comp_abs_chempot)

    # def test_pd_1d_draw(self):
    #     pdp = PDPlotter(self.pd_1d, show_unstable=10)
    #     pdp.show()

    def test_cpd_1d_draw(self):
        with self.assertRaises(NotImplementedError):
            self.cpd_1d.draw_diagram()

    def test_cpd_2d(self):
        self.assertEqual([Element.Ca, Element.Mg], self.cpd_2d.elements)
        self.assertEqual(2, self.cpd_2d.dim)
        self.assertEqual([[0.0, -5.0], [-5.0, 0.0]], self.cpd_2d.vertices)
#        self.assertEqual([[-5.0, 0.0], [0.0, -5.0]], self.cpd_2d.vertices)
        self.assertEqual("CaMg", self.cpd_2d.target_comp)
        self.assertEqual({'A': [0.0, -5.0], 'B': [-5.0, 0.0]},
                         self.cpd_2d.target_comp_chempot)
        self.assertEqual({'A': [-2.0, -6.0], 'B': [-7.0, -1.0]},
                         self.cpd_2d.target_comp_abs_chempot)

    @unittest.skipIf(DISABLE_DISPLAY_DIAGRAM, "not display chempotdiag")
    def test_cpd_2d_draw(self):
        self.cpd_2d.draw_diagram()

    def test_cpd_2d_draw_show(self):
        path = "vise.chempotdiag.chem_pot_diag.plt.show"
        with patch(path) as show_patch:
            self.cpd_2d.draw_diagram()
            assert show_patch.called

    def test_cpd_2d_draw_savefig(self):
        path = "vise.chempotdiag.chem_pot_diag.plt.savefig"
        with patch(path) as show_patch:
            self.cpd_2d.draw_diagram(filename="a.pdf")
            show_patch.assert_called_once_with("a.pdf")

    def test_cpd_3d(self):
        self.assertEqual([Element.Ca, Element.Mg, Element.O],
                         self.cpd_3d.elements)
        self.assertEqual(3, self.cpd_3d.dim)
        self.assertEqual([[-10.0, 0.0, 0.0], [-5.0, 0.0, -5.0],
                          [0.0, -10.0, 0.0], [0.0, -5.0, -5.0]],
                         self.cpd_3d.vertices)
        self.assertEqual("CaMgO", self.cpd_3d.target_comp)
        self.assertEqual({'A': [-10.0, 0.0, 0.0], 'B': [-5.0, 0.0, -5.0],
                          'C': [0.0, -10.0, 0.0], 'D': [0.0, -5.0, -5.0]},
                         self.cpd_3d.target_comp_chempot)

    @unittest.skipIf(DISABLE_DISPLAY_DIAGRAM, "not display chempotdiag")
    def test_cpd_3d_draw(self):
        self.cpd_3d.draw_diagram()

    def test_cpd_3d_unstable(self):
        self.assertEqual({'A': [-10.0, 0.0, 0.0], 'B': [0.0, -10.0, 0.0]},
                         self.cpd_3d_unstable.target_comp_chempot)

    @unittest.skipIf(DISABLE_DISPLAY_DIAGRAM, "not display chempotdiag")
    def test_cpd_3d_unstable_draw(self):
        self.cpd_3d_unstable.draw_diagram()

    def test_cpd_4d(self):
        self.assertEqual([Element.Sr, Element.Ca, Element.Mg, Element.O],
                         self.cpd_4d.elements)
        self.assertEqual([[0.0, -10.0, 0.0, 0.0], [0.0, 0.0, -10.0, 0.0],
                          [0.0, -5.0, 0.0, -5.0], [0.0, 0.0, -5.0, -5.0]],
                         self.cpd_4d.vertices)

    def test_pd_4d_plot(self):
        pdp = PDPlotter(self.pd_4d)
        pdp.show()

    def test_cpd_4d(self):
        # print(self.comp_cpd.elements)
        # print(self.comp_cpd.el_ref_list)
        # print(self.comp_cpd.dim)
        # print(self.comp_cpd.vertices)
        # print(self.comp_cpd.qhull_entries)
        # print(self.comp_cpd.comp_facets)
        # print(self.comp_cpd.target_comp)
        pdp = PDPlotter(self.comp_pd)
        pdp.show()

    def test_comp_cpd_4d(self):
#        self.comp_pd.draw_diagram()
        print(self.comp_cpd.target_comp_abs_chempot)
        self.comp_cpd.draw_diagram()


class TestSortCoords(ViseTest):
    def setUp(self) -> None:
        # x + 2y + 3z = 4
        self.coords = \
            np.array([[3, 2, -1], [-1, 2, 0], [-6, -1, 4], [1, -3, 3]])

    def test_sort_coords(self):
#        print(self.coords)
        print(sort_coords(self.coords))




# -*- coding: utf-8 -*-

import json
from typing import Optional
import yaml

from monty.json import MSONable
from monty.serialization import loadfn

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class PriorInfo(MSONable):
    """ Prior information for controlling parameters in DFT calculations """
    def __init__(self,
                 energy_per_atom: float = None,
                 band_gap: float = None,
                 total_magnetization: float = None,
                 data_source: str = None,
                 is_cluster: bool = None,
                 mag_threshold: float = 0.05,
                 band_gap_threshold: float = 0.1,
                 incar: dict = None,
                 **kwargs):
        """
            Args:
                energy_per_atom (float):
                    Energy per atom calculated in the data_source.
                band_gap (float):
                    Band gap calculated in the data_source.
                total_magnetization (float):
                    Total total_magnetization in the data_source.
                data_source (str):
                    The data source
                is_cluster (bool):
                    Whether the system is molecule or not.
                mag_threshold (float):
                    Threshold to judge if the system is magnetic.
                band_gap_threshold (float):
                    Threshold to judge if the system is metal.
                incar (dict):
                    Dict of INCAR flags to be set.
        """
        self.energy_per_atom = energy_per_atom
        self.band_gap = band_gap
        self.total_magnetization = total_magnetization
        self.data_source = data_source
        self.is_cluster = is_cluster
        self.mag_threshold = mag_threshold
        self.band_gap_threshold = band_gap_threshold
        self.incar = incar
        self.kwargs = kwargs

    def __repr__(self):
        outs = [f"energy_per_atom: {self.energy_per_atom}",
                f"band gap: {self.band_gap}",
                f"total magnetization: {self.total_magnetization}",
                f"data source: {self.data_source}",
                f"is cluster: {self.is_cluster}",
                f"incar flags {self.incar}"]
        return "\n".join(outs)

    def dump_yaml(self, filename: str = "prior_info.yaml") -> None:
        with open(filename, "w") as f:
            f.write(yaml.dump(self.as_dict()))

    @classmethod
    def load_yaml(cls, filename: str = "prior_info.yaml"):
        with open(filename, "r") as f:
            d = yaml.load(f, Loader=yaml.FullLoader)
        d["incar"] = d.get("incar", None)
        return cls.from_dict(d)

    def dump_json(self, filename: str = "prior_info.json") -> None:
        with open(filename, "w") as fw:
            json.dump(self.as_dict(), fw, indent=2)

    @classmethod
    def load_json(cls, filename: str = "prior_info.json"):
        return loadfn(filename)

    @property
    def is_magnetic(self) -> Optional[bool]:
        try:
            return self.total_magnetization > self.mag_threshold
        except TypeError:
            return

    @property
    def has_band_gap(self) -> bool:
        return self.band_gap > self.band_gap_threshold

    @property
    def is_metal(self) -> bool:
        return self.band_gap < self.band_gap_threshold



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
from typing import Union, List, Tuple, Dict, Any
from enum import Enum
import json
import math
import os
from pathlib import Path
import re
import yaml

from scipy.constants import speed_of_light, physical_constants, R, Avogadro

from monty.json import MSONable

from pymatgen.core.composition import Composition
from pymatgen.io.vasp.inputs import Poscar

from vise.config import REFERENCE_PRESSURE
from vise.util.error_classes import InvalidFileError
from vise.input_set.prior_info import PriorInfo

""" Data is from NIST Chemistry WebBook, SRD 69 """


class FundamentalFrequency(MSONable):

    def __init__(self, wave_number: float, degeneration: int):
        """

        Args:
            wave_number (float): frequency (cm^-1)
            degeneration (int): number of degeneration
        """
        self.wave_number = wave_number
        self.degeneration = degeneration

    def __str__(self):
        return f"wave_number: {self.wave_number} /cm, " \
               f"degeneration: {self.degeneration}"

    @property
    def zero_point_vibrational_energy(self) -> float:
        h = physical_constants["Planck constant in eV s"][0]
        sol = speed_of_light
        # h/2 * degeneration * wave_num(cm^-1) * 100(cm^-1->m^-1) * light speed
        return 0.5 * h * self.degeneration * self.wave_number * 100 * sol


class FundamentalFrequencies:

    def __init__(self,
                 frequencies: List[FundamentalFrequency],
                 reference: str):
        """
        Args:
            frequencies (list of FundamentalFrequency):
                Fundamental frequencies.
            reference (str):
                Citation of the data.
        """
        self.frequencies = frequencies
        self.reference = reference

    @classmethod
    def from_yaml(cls, file_path: str) -> "FundamentalFrequencies":
        """

        Args:
            file_path (str):

        Returns (FundamentalFrequencies):

        """

        with open(file_path, 'r') as fr:
            read_data = yaml.safe_load(fr)

        frequencies = \
            [FundamentalFrequency(f["wave_number"], f["degeneration"])
             for f in read_data["frequencies"]]
        try:
            reference = read_data["reference"]
        except KeyError:
            reference = None

        return cls(frequencies, reference)

    @property
    def zero_point_vibrational_energy(self) -> float:
        """
        Returns (float): E_ZPVE (eV)

        """
        return sum(f.zero_point_vibrational_energy for f in self.frequencies)


class AbstractThermodynamicsFunction(metaclass=ABCMeta):
    # TODO: implement abstract factory

    @abstractmethod
    def heat_capacity(self, temperature: float) -> float:
        pass

    @abstractmethod
    def standard_enthalpy(self, temperature: float) -> float:
        pass

    @abstractmethod
    def standard_entropy(self, temperature: float) -> float:
        pass

    @abstractmethod
    def r_ln_p_p0(self, pressure: float) -> float:
        pass


class ShomateParameters:
    """
        Parameters of Shomate Equations.
        Contains range of temperature like (100, 700).
    """

    def __init__(self,
                 temperature_range: Tuple[float, float],
                 a: float, b: float, c: float, d: float,
                 e: float, f: float, g: float, h: float):
        """

        Args:
            temperature_range (tuple): Unit is (K).
            a (float): parameter A.
            b (float): parameter B.
            c (float): parameter C.
            d (float): parameter D.
            e (float): parameter E.
            f (float): parameter F.
            g (float): parameter G.
            h (float): parameter H.
        """
        self.temperature_range = temperature_range
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.f = f
        self.g = g
        self.h = h

    def can_apply(self, temp: float) -> bool:
        return self.temperature_range[0] <= temp <= self.temperature_range[1]

    @property
    def min_temperature(self) -> float:
        return self.temperature_range[0]

    @property
    def max_temperature(self) -> float:
        return self.temperature_range[1]

    def __str__(self):
        outs = [f"Temperature_range: {self.temperature_range}\n",
                f"A: {self.a}\n", f"B: {self.b}\n", f"C: {self.c}\n",
                f"D: {self.d}\n", f"E: {self.e}\n", f"F: {self.f}\n",
                f"G: {self.g}\n", f"H: {self.h}\n"]
        return "\n".join(outs)


class ShomateThermodynamicsFunction(AbstractThermodynamicsFunction):
    """Shomate thermodynamics Function. """

    def __init__(self,
                 func_param_list: List[ShomateParameters],
                 enthalpy_zero: float,
                 reference_pressure: float = REFERENCE_PRESSURE):
        """
        Args:
            func_param_list (list): parameters of functions.
                                    Must contain _Shomate_parameters.
            enthalpy_zero (float): enthalpy (J/mol) at T=0K,
                                   p=reference_pressure
            reference_pressure (float): reference_pressure (Pa)

        """
        self._params = func_param_list
        self.enthalpy_zero = enthalpy_zero
        self.reference_pressure = reference_pressure

    @classmethod
    def from_nist_table(cls, file_path: str) -> "ShomateThermodynamicsFunction":
        """
        shomate_nist.dat is written e.g.
        Enthalpy_zero   -8.825
        Temperature (K)	298. - 6000.
        A	31.44510
        B	8.413831
        C	-2.778850
        D	0.218104
        E	-0.211175
        F	-10.43260
        G	237.2770
        H	0.000000
        Reference	Chase, 1998
        Comment	Data last reviewed in June, 1982

        where Enthalpy_zero is kJ/mol

        Args:
            file_path (str):

        Returns:
            ShomateThermodynamicsFunction object
        """
        # TODO: unit of file is confusing
        # Is there better expression with regular expression?
        temperature_ranges = []
        param_dicts = []
        enthalpy_zero = None
        with open(file_path, "r") as fr:
            for line in fr:
                if enthalpy_zero is None:
                    is_tmp_line = \
                        bool(re.match(r"Enthalpy_zero", line))
                    if is_tmp_line:
                        matched = re.findall(r"[\-\d\.]+", line)
                        enthalpy_zero = float(matched[0]) / 1000
                    continue

                if not temperature_ranges:
                    is_tmp_line = \
                        bool(re.match(r"Temperature\s+\(K\)\s+", line))
                    if is_tmp_line:
                        matched = re.findall(r"[\d\.]+", line)
                        for i in range(int(len(matched)/2)):
                            temperature_ranges.append(
                                (float(matched[2*i]), float(matched[2*i+1])))
                        param_dicts = \
                            [dict() for _ in range(len(temperature_ranges))]
                else:
                    symbol_match = re.search(r"\s*([A-H])\s+", line)
                    if symbol_match:
                        symbol = symbol_match.groups()[0].lower()
                        values = re.findall(r"[\-\d\.]+", line)
                        if len(values) != len(temperature_ranges):
                            raise ValueError(f"Number of temperature_ranges"
                                             f"({len(temperature_ranges)})"
                                             f"and parameters ({line})"
                                             f"is not consistent.")
                        for i in range(len(values)):
                            param_dicts[i][symbol] = float(values[i])
        return cls([ShomateParameters(tr, **pd)
                    for tr, pd in zip(temperature_ranges, param_dicts)],
                   enthalpy_zero)

    def params(self, temperature: float) -> ShomateParameters:
        """Return parameters of function at temperature like {a:1.1, b:2.2, ...}

        Args:
            temperature:
        Returns:

        """
        params = None
        for p in self._params:
            if p.can_apply(temperature):
                params = p
        if params is None:
            raise ValueError(f"Temperature {temperature} is out of temperature "
                             f"range to apply")
        return params

    @property
    def temperature_range(self) -> Tuple[float, float]:
        return self.min_temperature, self.max_temperature

    @property
    def max_temperature(self) -> float:
        max_data = max(self._params, key=lambda x: x.max_temperature)
        return max_data.temperature_range[1]

    @property
    def min_temperature(self) -> float:
        min_data = min(self._params, key=lambda x: x.min_temperature)
        return min_data.temperature_range[0]

    def heat_capacity(self, temperature: float) -> float:
        """

        Args:
            temperature (float): (K)

        Returns (float): J/(K mol)

        """
        params = self.params(temperature)
        t = temperature / 1000
        a, b, c, d, e = params.a, params.b, params.c, params.d, params.e
        return a + b * t + c * t ** 2 + d * t ** 3 + e / (t ** 2)

    def standard_enthalpy(self, temperature: float) -> float:
        """

        Args:
            temperature (float): (K)

        Returns (float): J/mol

        """
        params = self.params(temperature)
        t = temperature / 1000
        a, b, c, d, e, f, h = \
            params.a, params.b, params.c, params.d, params.e, params.f, params.h
        enthalpy_kj = (a * t + b * t ** 2 / 2 + c * t ** 3 / 3
                       + d * t ** 4 / 4 - e / t + f - h)
        return enthalpy_kj / 1000

    def standard_entropy(self, temperature: float) -> float:
        """

        Args:
            temperature (float): (K)

        Returns (float): J/(mol K)

        """
        params = self.params(temperature)
        t = temperature / 1000
        a, b, c, d, e, g = \
            params.a, params.b, params.c, params.d, params.e, params.g
        return (a * math.log(t) + b * t + c * t ** 2 / 2 + d * t ** 3 / 3
                - e / (2 * t ** 2) + g)

    def r_ln_p_p0(self, pressure: float) -> float:
        """ R ln(p/p0)
        Args:
            pressure (float): Pa

        Returns (float): Rln(p/p0), (J / (mol K))

        """
        return R * math.log(pressure / self.reference_pressure)


class Gas(Enum):
    F2 = "F2"
    H2 = "H2"
    H2O = "H2O"
    N2 = "N2"
    NH3 = "NH3"
    NO2 = "NO2"
    O2 = "O2"
    P4 = "P4"

    def __str__(self):
        return self.name

    def __init__(self, formula: Union[str, Composition]):
        self.formula = str(formula)
        dirname = Path(__file__).parent / "molecules" / self.formula

        try:
            self.properties = PriorInfo.load_json(dirname / "prior_info.yaml")
        except:
            pass
        self.thermodynamics_function = ShomateThermodynamicsFunction.\
            from_nist_table(dirname / "shomate_nist.dat")
        self.fundamental_frequencies = FundamentalFrequencies.\
            from_yaml(dirname / "fundamental_frequencies.yaml")

    @classmethod
    def name_list(cls):
        return ', '.join([e.value for e in cls])

    @property
    def temperature_range(self) -> Tuple[float, float]:
        """Temperature range that can be applied. """
        return self.thermodynamics_function.temperature_range

    @property
    def min_temperature(self) -> float:
        """Minimum temperature that can be applied. """
        return self.thermodynamics_function.min_temperature

    @property
    def max_temperature(self) -> float:
        """Maximum temperature that can be applied. """
        return self.thermodynamics_function.max_temperature

    def r_ln_p_p0_j_mol(self, pressure: float) -> float:
        """Rln(p/p0), (J / (mol K)) """
        return self.thermodynamics_function.r_ln_p_p0(pressure)
    
    def r_ln_p_p0_ev_per_atom(self, pressure: float) -> float:
        """Rln(p/p0), (eV / atom) """
        j2ev = physical_constants["joule-electron volt relationship"][0]
        return j2ev * self.r_ln_p_p0_j_mol(pressure) / (self.n_atom * Avogadro)

    @property
    def n_atom(self) -> int:
        return Composition(self.formula).num_atoms

    @property
    def zero_point_vibrational_energy(self) -> float:
        """Energy of vibrational energy at zero point (eV/atom) """
        zpve = self.fundamental_frequencies.zero_point_vibrational_energy
        return zpve / self.n_atom

    def heat_capacity(self, temperature: float) -> float:
        """Returns (float): Heat capacity C_p (J/mol*K). """
        return self.thermodynamics_function.heat_capacity(temperature)

    def standard_enthalpy(self, temperature: float) -> float:
        """Standard enthalpy (kJ/mol) from room temperature (T=298.15(K)) """
        return self.thermodynamics_function.standard_enthalpy(temperature)

    def standard_entropy(self, temperature: float) -> float:
        """Standard entropy (J/mol*K). """
        return self.thermodynamics_function.standard_entropy(temperature)

    def vib_rot_term(self, temperature: float) -> float:
        """G(T, P0) - H(T=0K, P0) (eV/mol) """
        h = self.standard_enthalpy(temperature)
        s = self.standard_entropy(temperature)
        h_0 = self.thermodynamics_function.enthalpy_zero
        val_joule = h - temperature * s - h_0
        joule_to_ev = physical_constants["joule-electron volt relationship"][0]
        return val_joule * joule_to_ev / (self.n_atom * Avogadro)

    def free_e_shift(self,
                     temperature: float = 0,
                     pressure: float = 1e+5) -> float:
        """Free energy shift (eV/atom) """
        if abs(temperature) < 1e-12:
            return 0.0

        vib_rot_term = self.vib_rot_term(temperature)
        trans_term = temperature * self.r_ln_p_p0_ev_per_atom(pressure)
        return vib_rot_term + trans_term

    def energy_shift(self,
                     temperature: float = 0,  
                     pressure: float = 1e+5) -> float:
        """Free energy shift including zero point vibration (eV/atom) """
        if abs(temperature) < 1e-12:
            return self.zero_point_vibrational_energy

        zero_point_vib = self.zero_point_vibrational_energy
        free_e_shift = self.free_e_shift(temperature, pressure)
        return zero_point_vib + free_e_shift
